import hashlib
import json
import subprocess
import uuid
from dataclasses import dataclass
from tempfile import NamedTemporaryFile

import gnupg
import pytest
import requests
from pulpcore.client.pulp_rpm import (
    AcsRpmApi,
    ContentAdvisoriesApi,
    ContentDistributionTreesApi,
    ContentModulemdDefaultsApi,
    ContentModulemdObsoletesApi,
    ContentModulemdsApi,
    ContentPackagecategoriesApi,
    ContentPackagegroupsApi,
    ContentPackagelangpacksApi,
    ContentPackagesApi,
    RemotesUlnApi,
    RpmCompsApi,
    RpmCopyApi,
    RpmRepositorySyncURL,
)

from pulp_rpm.tests.functional.constants import (
    BASE_TEST_JSON,
    RPM_KICKSTART_FIXTURE_URL,
    RPM_MODULAR_FIXTURE_URL,
    RPM_SIGNED_FIXTURE_URL,
    RPM_SIGNED_URL,
)
from pulp_rpm.tests.functional.utils import init_signed_repo_configuration


@pytest.fixture(scope="session")
def rpm_ulnremote_api(rpm_client):
    """Fixture for RPM remote API."""
    return RemotesUlnApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_acs_api(rpm_client):
    """Fixture for RPM alternate content source API."""
    return AcsRpmApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_package_api(rpm_client):
    """Fixture for RPM distribution API."""
    return ContentPackagesApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_advisory_api(rpm_client):
    """Fixture for RPM distribution API."""
    return ContentAdvisoriesApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_package_category_api(rpm_client):
    """Fixture for RPM distribution API."""
    return ContentPackagecategoriesApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_package_groups_api(rpm_client):
    """Fixture for RPM distribution API."""
    return ContentPackagegroupsApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_package_lang_packs_api(rpm_client):
    """Fixture for RPM distribution API."""
    return ContentPackagelangpacksApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_modulemd_api(rpm_client):
    """Fixture for RPM Modulemd API."""
    return ContentModulemdsApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_modulemd_defaults_api(rpm_client):
    """Fixture for RPM ModulemdDefault API."""
    return ContentModulemdDefaultsApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_modulemd_obsoletes_api(rpm_client):
    """Fixture for RPM ModulemdObsolete API."""
    return ContentModulemdObsoletesApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_comps_api(rpm_client):
    """Fixture for RPM Comps API."""
    return RpmCompsApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_content_distribution_trees_api(rpm_client):
    return ContentDistributionTreesApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_copy_api(rpm_client):
    return RpmCopyApi(rpm_client)


@pytest.fixture
def signed_artifact(pulpcore_bindings, tmp_path):
    data = requests.get(RPM_SIGNED_URL).content
    artifacts = pulpcore_bindings.ArtifactsApi.list(
        sha256=hashlib.sha256(data).hexdigest(), limit=1
    )
    try:
        return artifacts.results[0]
    except IndexError:
        pass

    temp_file = tmp_path / str(uuid.uuid4())
    temp_file.write_bytes(data)
    return pulpcore_bindings.ArtifactsApi.create(str(temp_file))


@pytest.fixture
def rpm_artifact_factory(pulpcore_bindings, gen_object_with_cleanup, pulp_domain_enabled, tmp_path):
    """Return an artifact created from uploading an RPM file."""

    def _rpm_artifact_factory(url=RPM_SIGNED_URL, pulp_domain=None):
        temp_file = tmp_path / str(uuid.uuid4())
        temp_file.write_bytes(requests.get(url).content)
        kwargs = {}
        if pulp_domain:
            if not pulp_domain_enabled:
                raise RuntimeError("Server does not have domains enabled.")
            kwargs["pulp_domain"] = pulp_domain
        return gen_object_with_cleanup(pulpcore_bindings.ArtifactsApi, str(temp_file), **kwargs)

    return _rpm_artifact_factory


@pytest.fixture
def rpm_package_factory(
    gen_object_with_cleanup,
    pulp_domain_enabled,
    rpm_package_api,
):
    """Return a Package created from uploading an RPM file."""

    def _rpm_package_factory(url=RPM_SIGNED_URL, pulp_domain=None):
        with NamedTemporaryFile() as file_to_upload:
            file_to_upload.write(requests.get(url).content)
            file_to_upload.flush()
            upload_attrs = {"file": file_to_upload.name}

            kwargs = {}
            if pulp_domain:
                if not pulp_domain_enabled:
                    raise RuntimeError("Server does not have domains enabled.")
                kwargs["pulp_domain"] = pulp_domain

            return gen_object_with_cleanup(rpm_package_api, **upload_attrs, **kwargs)

    return _rpm_package_factory


@pytest.fixture(scope="class")
def rpm_unsigned_repo_immediate(init_and_sync):
    repo, _ = init_and_sync()
    return repo


@pytest.fixture(scope="class")
def rpm_unsigned_repo_on_demand(init_and_sync):
    repo, _ = init_and_sync(policy="on_demand")
    return repo


@pytest.fixture(scope="class")
def rpm_modular_repo_on_demand(init_and_sync):
    repo, _ = init_and_sync(url=RPM_MODULAR_FIXTURE_URL, policy="on_demand")
    return repo


@pytest.fixture(scope="class")
def rpm_kickstart_repo_immediate(init_and_sync):
    repo, _ = init_and_sync(url=RPM_KICKSTART_FIXTURE_URL)
    return repo


@pytest.fixture(scope="session")
def rpm_metadata_signing_service(pulpcore_bindings):
    results = pulpcore_bindings.SigningServicesApi.list(name="sign-metadata")
    signing_service = None
    if results.count == 0:
        result = init_signed_repo_configuration()
        if result.returncode == 0:
            results = pulpcore_bindings.SigningServicesApi.list(name="sign-metadata")
    if results.count == 1:
        signing_service = results.results[0]

    return signing_service


@pytest.fixture
def upload_wrong_file_type(rpm_advisory_api):
    def _upload(remote_path):
        with NamedTemporaryFile() as file_to_upload:
            file_to_upload.write(requests.get(remote_path).content)
            file_to_upload.flush()
            upload_attrs = {"file": file_to_upload.name}
            return rpm_advisory_api.create(**upload_attrs)

    return _upload


@pytest.fixture
def upload_advisory_factory(
    add_to_cleanup,
    monitor_task,
    pulp_domain_enabled,
    rpm_advisory_api,
):
    """Upload advisory from a json file, return advisory, vers_href, and id-used."""

    def _upload_advisory_factory(
        advisory=BASE_TEST_JSON, repository=None, pulp_domain=None, set_id=False, use_id=None
    ):
        kwargs = {}
        if pulp_domain:
            if not pulp_domain_enabled:
                raise RuntimeError("Server does not have domains enabled.")
            kwargs["pulp_domain"] = pulp_domain

        with NamedTemporaryFile("w+") as file_to_upload:
            json_advisory = json.loads(advisory)
            if set_id or use_id:
                json_advisory["id"] = use_id if use_id else str(uuid.uuid4())
            used_id = json_advisory["id"]
            json.dump(json_advisory, file_to_upload)
            file_to_upload.flush()
            upload_attrs = {"file": file_to_upload.name}
            if repository:
                upload_attrs["repository"] = repository.pulp_href
            file_to_upload.flush()
            response = rpm_advisory_api.create(**upload_attrs, **kwargs)

        task_rslt = monitor_task(response.task)
        if repository:
            assert 2 == len(task_rslt.created_resources)
        else:
            assert 1 == len(task_rslt.created_resources)

        vers_href = None
        advisory_href = None
        for rsrc in task_rslt.created_resources:
            if "versions" in rsrc:
                vers_href = rsrc
            elif "advisories" in rsrc:
                advisory_href = rsrc
        assert advisory_href

        add_to_cleanup(rpm_advisory_api, advisory_href)
        entity = rpm_advisory_api.read(advisory_href)
        return entity, vers_href, used_id

    return _upload_advisory_factory


@pytest.fixture
def assert_uploaded_advisory(rpm_advisory_api):
    """List advisories for a given version-href, and assert that the specified ID is therein."""

    def _from_results(advisory_id, vers_href):
        advisories = rpm_advisory_api.list(id=advisory_id, repository_version=vers_href)
        assert 1 == len(advisories.results)
        return advisories.results[0].pulp_href, vers_href

    return _from_results


@pytest.fixture
def setup_domain(
    gen_object_with_cleanup, pulpcore_bindings, rpm_rpmremote_api, rpm_repository_api, monitor_task
):
    def _setup_domain(sync=True, url=RPM_SIGNED_FIXTURE_URL, pulp_domain=None):
        if not pulp_domain:
            body = {
                "name": str(uuid.uuid4()),
                "storage_class": "pulpcore.app.models.storage.FileSystem",
                "storage_settings": {"MEDIA_ROOT": "/var/lib/pulp/media/"},
            }
            pulp_domain = gen_object_with_cleanup(pulpcore_bindings.DomainsApi, body)

        remote = gen_object_with_cleanup(
            rpm_rpmremote_api, {"name": str(uuid.uuid4()), "url": url}, pulp_domain=pulp_domain.name
        )
        src = gen_object_with_cleanup(
            rpm_repository_api,
            {"name": str(uuid.uuid4()), "remote": remote.pulp_href},
            pulp_domain=pulp_domain.name,
        )

        if sync:
            sync_url = RpmRepositorySyncURL()
            monitor_task(rpm_repository_api.sync(src.pulp_href, sync_url).task)
            src = rpm_repository_api.read(src.pulp_href)

        dest = gen_object_with_cleanup(
            rpm_repository_api, {"name": str(uuid.uuid4())}, pulp_domain=pulp_domain.name
        )
        return pulp_domain, remote, src, dest

    return _setup_domain


@pytest.fixture
def cleanup_domains(pulpcore_bindings, monitor_task, rpm_repository_api):
    def _cleanup_domains(
        domains,
        content_api_client=None,
        cleanup_repositories=False,
        repository_api_client=rpm_repository_api,
    ):
        for domain in domains:
            # clean up each domain specified
            if domain:
                if cleanup_repositories:
                    # Delete repos from the domain
                    for repo in repository_api_client.list(pulp_domain=domain.name).results:
                        monitor_task(repository_api_client.delete(repo.pulp_href).task)
                # Let orphan-cleanup reap the resulting abandoned content
                monitor_task(
                    pulpcore_bindings.OrphansCleanupApi.cleanup(
                        {"orphan_protection_time": 0}, pulp_domain=domain.name
                    ).task
                )

        if content_api_client:
            # IF we have a client, check that each domain is empty of that kind-of entity
            for domain in domains:
                if domain:
                    assert content_api_client.list(pulp_domain=domain.name).count == 0

    return _cleanup_domains


# package signing

SIGNING_SCRIPT_STRING = r"""#!/usr/bin/env bash
# Rpm configuration:
#     GPG_HOME: gpg home directory
#     GPG_NAME: gpg user identity
#     GPG_BIN: gpg binary path

FILE_PATH=$1
GPG_HOME=HOMEDIRHERE
GPG_BIN=/usr/bin/gpg

# user id can be specified by a fingerprint:
# see https://www.gnupg.org/documentation/manuals/gnupg/Specify-a-User-ID.html
GPG_NAME="${PULP_SIGNING_KEY_FINGERPRINT}"

# Sign the package
rpm \
    --define "_signature gpg" \
    --define "_gpg_path ${GPG_HOME}" \
    --define "_gpg_name ${GPG_NAME}" \
    --define "_gpgbin ${GPG_BIN}" \
    --addsign "${FILE_PATH}" 1> /dev/null

# Check the exit status
STATUS=$?
if [[ ${STATUS} -eq 0 ]]; then
   echo {\"rpm_package\": \"${FILE_PATH}\"}
else
   exit ${STATUS}
fi
"""


@pytest.fixture(scope="session")
def signing_script_path(signing_script_temp_dir, signing_gpg_homedir_path):
    signing_script_file = signing_script_temp_dir / "sign-rpm-package.sh"
    signing_script_file.write_text(
        SIGNING_SCRIPT_STRING.replace("HOMEDIRHERE", str(signing_gpg_homedir_path))
    )

    signing_script_file.chmod(0o755)

    return signing_script_file


@pytest.fixture(scope="session")
def signing_script_temp_dir(tmp_path_factory):
    return tmp_path_factory.mktemp("sigining_script_dir")


@pytest.fixture(scope="session")
def signing_gpg_homedir_path(tmp_path_factory):
    return tmp_path_factory.mktemp("gpghome")


@pytest.fixture
def sign_with_rpm_package_signing_service(signing_script_path, signing_gpg_metadata):
    """
    Runs the test signing script manually, locally, and returns the signature file produced.
    """

    def _sign_with_rpm_package_signing_service(filename):
        env = {"PULP_SIGNING_KEY_FINGERPRINT": signing_gpg_metadata[1]}
        cmd = (signing_script_path, filename)
        completed_process = subprocess.run(
            cmd,
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        if completed_process.returncode != 0:
            raise RuntimeError(str(completed_process.stderr))

        try:
            return_value = json.loads(completed_process.stdout)
        except json.JSONDecodeError:
            raise RuntimeError("The signing script did not return valid JSON!")

        return return_value

    return _sign_with_rpm_package_signing_service


@dataclass
class GPGMetadata:
    public_key: str
    fingerprint: str
    keyid: str


@pytest.fixture(scope="session")
def signing_gpg_metadata2(signing_gpg_homedir_path) -> tuple[gnupg.GPG, list[GPGMetadata]]:
    """
    A fixture that returns a GPG instance and related metadata (i.e., fingerprint, keyid).
    """
    PRIVATE_KEY_URLS = (
        "https://raw.githubusercontent.com/pulp/pulp-fixtures/master/common/GPG-PRIVATE-KEY-fixture-signing",  # noqa: E501
        "https://raw.githubusercontent.com/pulp/pulp-fixtures/master/common/GPG-PRIVATE-KEY-pulp-qe",  # noqa: E501
    )

    gpg = gnupg.GPG(gnupghome=signing_gpg_homedir_path)
    keys = []
    for privatekey_url in PRIVATE_KEY_URLS:
        response_private = requests.get(privatekey_url)
        response_private.raise_for_status()

        gpg.import_keys(response_private.content)
        key_info = gpg.list_keys()[-1]
        gpg_md = GPGMetadata(
            fingerprint=key_info["fingerprint"],
            keyid=key_info["keyid"],
            public_key=gpg.export_keys(key_info["keyid"]),
        )
        gpg.trust_keys(gpg_md.fingerprint, "TRUST_ULTIMATE")
        keys.append(gpg_md)

    return (gpg, keys)


@pytest.fixture(scope="session")
def signing_gpg_metadata(signing_gpg_homedir_path):
    """
    A fixture that returns a GPG instance and related metadata (i.e., fingerprint, keyid).
    """
    PRIVATE_KEY_URL = "https://raw.githubusercontent.com/pulp/pulp-fixtures/master/common/GPG-PRIVATE-KEY-fixture-signing"  # noqa: E501

    response_private = requests.get(PRIVATE_KEY_URL)
    response_private.raise_for_status()

    gpg = gnupg.GPG(gnupghome=signing_gpg_homedir_path)
    gpg.import_keys(response_private.content)

    fingerprint = gpg.list_keys()[0]["fingerprint"]
    keyid = gpg.list_keys()[0]["keyid"]

    gpg.trust_keys(fingerprint, "TRUST_ULTIMATE")

    return gpg, fingerprint, keyid


@pytest.fixture(scope="session")
def pulp_trusted_public_key(signing_gpg_metadata):
    """Fixture to extract the ascii armored trusted public test key."""
    gpg, _, keyid = signing_gpg_metadata
    return gpg.export_keys([keyid])


@pytest.fixture(scope="session")
def pulp_trusted_public_key_fingerprint(signing_gpg_metadata):
    """Fixture to extract the ascii armored trusted public test keys fingerprint."""
    return signing_gpg_metadata[1]


@pytest.fixture(scope="session")
def _rpm_package_signing_service_name(
    bindings_cfg,
    signing_script_path,
    signing_gpg_metadata,
    signing_gpg_homedir_path,
    pytestconfig,
):
    service_name = str(uuid.uuid4())
    gpg, fingerprint, keyid = signing_gpg_metadata

    cmd = (
        "pulpcore-manager",
        "add-signing-service",
        service_name,
        str(signing_script_path),
        fingerprint,
        "--class",
        "rpm:RpmPackageSigningService",
        "--gnupghome",
        str(signing_gpg_homedir_path),
    )
    completed_process = subprocess.run(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    assert completed_process.returncode == 0

    yield service_name

    cmd = (
        "pulpcore-manager",
        "remove-signing-service",
        service_name,
        "--class",
        "rpm:RpmPackageSigningService",
    )
    subprocess.run(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )


@pytest.fixture
def rpm_package_signing_service(_rpm_package_signing_service_name, pulpcore_bindings):
    return pulpcore_bindings.SigningServicesApi.list(
        name=_rpm_package_signing_service_name
    ).results[0]



# Generated by Django 2.2.16 on 2020-09-18 13:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0022_add_collections_related_name'),
    ]

    operations = [
        migrations.AlterField(
            model_name='distributiontree',
            name='base_product_short',
            field=models.CharField(max_length=50, null=True),
        ),
        migrations.AlterField(
            model_name='distributiontree',
            name='release_short',
            field=models.CharField(max_length=50),
        ),
        migrations.AlterField(
            model_name='image',
            name='name',
            field=models.CharField(max_length=50),
        ),
    ]



# Generated by Django 4.2.1 on 2023-05-12 13:24

from django.db import migrations, models
import django.db.models.deletion
import pulpcore.app.models.base


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0106_alter_artifactdistribution_distribution_ptr_and_more'),
        ('rpm', '0049_profiles_fix'),
    ]

    operations = [
        migrations.AlterField(
            model_name='addon',
            name='pulp_id',
            field=models.UUIDField(default=pulpcore.app.models.base.pulp_uuid, editable=False, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='checksum',
            name='pulp_id',
            field=models.UUIDField(default=pulpcore.app.models.base.pulp_uuid, editable=False, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='distributiontree',
            name='content_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content'),
        ),
        migrations.AlterField(
            model_name='image',
            name='pulp_id',
            field=models.UUIDField(default=pulpcore.app.models.base.pulp_uuid, editable=False, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='modulemd',
            name='content_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content'),
        ),
        migrations.AlterField(
            model_name='modulemd',
            name='packages',
            field=models.ManyToManyField(to='rpm.package'),
        ),
        migrations.AlterField(
            model_name='modulemddefaults',
            name='content_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content'),
        ),
        migrations.AlterField(
            model_name='modulemdobsolete',
            name='content_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content'),
        ),
        migrations.AlterField(
            model_name='package',
            name='content_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content'),
        ),
        migrations.AlterField(
            model_name='packagecategory',
            name='content_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content'),
        ),
        migrations.AlterField(
            model_name='packageenvironment',
            name='content_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content'),
        ),
        migrations.AlterField(
            model_name='packagegroup',
            name='content_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content'),
        ),
        migrations.AlterField(
            model_name='packagelangpacks',
            name='content_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content'),
        ),
        migrations.AlterField(
            model_name='repometadatafile',
            name='content_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content'),
        ),
        migrations.AlterField(
            model_name='rpmalternatecontentsource',
            name='alternatecontentsource_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.alternatecontentsource'),
        ),
        migrations.AlterField(
            model_name='rpmdistribution',
            name='distribution_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.distribution'),
        ),
        migrations.AlterField(
            model_name='rpmpublication',
            name='publication_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.publication'),
        ),
        migrations.AlterField(
            model_name='rpmremote',
            name='remote_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.remote'),
        ),
        migrations.AlterField(
            model_name='rpmrepository',
            name='metadata_signing_service',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.asciiarmoreddetachedsigningservice'),
        ),
        migrations.AlterField(
            model_name='rpmrepository',
            name='repository_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.repository'),
        ),
        migrations.AlterField(
            model_name='ulnremote',
            name='remote_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.remote'),
        ),
        migrations.AlterField(
            model_name='updatecollection',
            name='pulp_id',
            field=models.UUIDField(default=pulpcore.app.models.base.pulp_uuid, editable=False, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='updatecollectionpackage',
            name='pulp_id',
            field=models.UUIDField(default=pulpcore.app.models.base.pulp_uuid, editable=False, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='updaterecord',
            name='content_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.content'),
        ),
        migrations.AlterField(
            model_name='updatereference',
            name='pulp_id',
            field=models.UUIDField(default=pulpcore.app.models.base.pulp_uuid, editable=False, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='variant',
            name='pulp_id',
            field=models.UUIDField(default=pulpcore.app.models.base.pulp_uuid, editable=False, primary_key=True, serialize=False),
        ),
    ]



# Generated by Django 4.2.10 on 2024-04-25 16:39

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("rpm", "0061_fix_modulemd_defaults_digest"),
    ]

    operations = [
        migrations.CreateModel(
            name="RpmPackageSigningService",
            fields=[
                (
                    "signingservice_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="core.signingservice",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("core.signingservice",),
        ),
        migrations.AddField(
            model_name="rpmrepository",
            name="package_signing_fingerprint",
            field=models.TextField(max_length=40, null=True),
        ),
        migrations.AddField(
            model_name="rpmrepository",
            name="package_signing_service",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="rpm.rpmpackagesigningservice",
            ),
        ),
    ]



import uuid

import pytest
from urllib.parse import urljoin
import requests

from pulpcore.client.pulp_rpm import (
    ApiClient as RpmApiClient,
    ContentRepoMetadataFilesApi,
    DistributionsRpmApi,
    PublicationsRpmApi,
    RemotesRpmApi,
    RepositoriesRpmApi,
    RepositoriesRpmVersionsApi,
    RpmPruneApi,
    RpmRepositorySyncURL,
)

from pulp_rpm.tests.functional.constants import RPM_UNSIGNED_FIXTURE_URL, RPM_CONTENT_NAMES


@pytest.fixture(scope="session")
def rpm_publication_api(rpm_client):
    """Fixture for RPM publication API."""
    return PublicationsRpmApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_repository_version_api(rpm_client):
    """Fixture for the RPM repository versions API."""
    return RepositoriesRpmVersionsApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_distribution_api(rpm_client):
    """Fixture for RPM distribution API."""
    return DistributionsRpmApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_prune_api(rpm_client):
    """Fixture for RPM Prune API."""
    return RpmPruneApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_client(bindings_cfg):
    """Fixture for RPM client."""
    return RpmApiClient(bindings_cfg)


@pytest.fixture(scope="session")
def rpm_content_repometadata_files_api(rpm_client):
    return ContentRepoMetadataFilesApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_rpmremote_api(rpm_client):
    """Fixture for RPM remote API."""
    return RemotesRpmApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_repository_api(rpm_client):
    """Fixture for RPM repositories API."""
    return RepositoriesRpmApi(rpm_client)


@pytest.fixture(scope="class")
def rpm_repository_factory(rpm_repository_api, gen_object_with_cleanup):
    """A factory to generate an RPM Repository with auto-deletion after the test run."""

    def _rpm_repository_factory(pulp_domain=None, **body):
        data = {"name": str(uuid.uuid4())}
        data.update(body)
        kwargs = {}
        if pulp_domain:
            kwargs["pulp_domain"] = pulp_domain
        return gen_object_with_cleanup(rpm_repository_api, data, **kwargs)

    return _rpm_repository_factory


@pytest.fixture(scope="class")
def rpm_rpmremote_factory(rpm_rpmremote_api, gen_object_with_cleanup):
    """A factory to generate an RPM Remote with auto-deletion after the test run."""

    def _rpm_rpmremote_factory(
        *, url=RPM_UNSIGNED_FIXTURE_URL, policy="immediate", pulp_domain=None, **body
    ):
        data = {"url": url, "policy": policy, "name": str(uuid.uuid4())}
        data.update(body)
        kwargs = {}
        if pulp_domain:
            kwargs["pulp_domain"] = pulp_domain
        return gen_object_with_cleanup(rpm_rpmremote_api, data, **kwargs)

    return _rpm_rpmremote_factory


@pytest.fixture(scope="class")
def rpm_distribution_factory(rpm_distribution_api, gen_object_with_cleanup):
    """A factory to generate an RPM Distribution with auto-deletion after the test run."""

    def _rpm_distribution_factory(pulp_domain=None, **body):
        data = {"base_path": str(uuid.uuid4()), "name": str(uuid.uuid4())}
        data.update(body)
        kwargs = {}
        if pulp_domain:
            kwargs["pulp_domain"] = pulp_domain
        return gen_object_with_cleanup(rpm_distribution_api, data, **kwargs)

    return _rpm_distribution_factory


@pytest.fixture(scope="class")
def rpm_publication_factory(rpm_publication_api, gen_object_with_cleanup):
    """A factory to generate an RPM Publication with auto-deletion after the test run."""

    def _rpm_publication_factory(pulp_domain=None, **body):
        # XOR check on repository and repository_version
        assert bool("repository" in body) ^ bool("repository_version" in body)
        kwargs = {}
        if pulp_domain:
            kwargs["pulp_domain"] = pulp_domain
        return gen_object_with_cleanup(rpm_publication_api, body, **kwargs)

    return _rpm_publication_factory


@pytest.fixture(scope="class")
def init_and_sync(rpm_repository_factory, rpm_repository_api, rpm_rpmremote_factory, monitor_task):
    """Initialize a new repository and remote and sync the content from the passed URL."""

    def _init_and_sync(
        repository=None,
        remote=None,
        url=RPM_UNSIGNED_FIXTURE_URL,
        policy="immediate",
        sync_policy="additive",
        skip_types=None,
        optimize=True,
        return_task=False,
    ):
        if repository is None:
            repository = rpm_repository_factory()
        if remote is None:
            remote = rpm_rpmremote_factory(url=url, policy=policy)

        repository_sync_data = RpmRepositorySyncURL(
            remote=remote.pulp_href,
            sync_policy=sync_policy,
            skip_types=skip_types,
            optimize=optimize,
        )
        sync_response = rpm_repository_api.sync(repository.pulp_href, repository_sync_data)
        task = monitor_task(sync_response.task)

        repository = rpm_repository_api.read(repository.pulp_href)
        return (repository, remote) if not return_task else (repository, remote, task)

    return _init_and_sync


class BaseURLSession(requests.Session):
    def __init__(self, base_url, *args, **kwargs):
        self.base_url = base_url
        super().__init__(*args, **kwargs)

    def request(self, method, url, **kwargs):
        return super().request(method, urljoin(self.base_url, url), **kwargs)


@pytest.fixture(scope="module")
def pulp_requests(bindings_cfg):
    """Uses requests lib to issue an http request to pulp server using pulp_href.

    Example:
        >>> response = pulp_requests.get("/pulp/api/v3/.../?repository_version=...")
        >>> type(response)
        requests.Response
    """
    with BaseURLSession(bindings_cfg.host) as session:
        session.auth = (bindings_cfg.username, bindings_cfg.password)
        yield session


@pytest.fixture
def get_content_summary(rpm_repository_version_api):
    """A fixture that fetches the content summary from a repository."""

    def _get_content_summary(repo, version_href=None, dump=True):
        """Fetches the content summary from a given repository.

        Args:
            repo: The repository where the content is fetched from.
            version_href: The repository version from where the content should be fetched from.
                Default: latest repository version.
            dump: If true, return a dumped dictionary with convenient filters (default).
                Otherwise, return the response object.

        Returns:
            The content summary of the repository.
        """
        version_href = version_href or repo.latest_version_href
        if version_href is None:
            return {}
        content_summary = rpm_repository_version_api.read(version_href).content_summary
        if not dump:
            return content_summary
        else:
            # removes the hrefs, which is may get in the way of data comparision
            # https://docs.pydantic.dev/latest/concepts/serialization/#pickledumpsmodel
            exclude_fields = {"__all__": {"__all__": {"href"}}}
            return content_summary.model_dump(exclude=exclude_fields)

    return _get_content_summary


@pytest.fixture
def get_content(
    rpm_repository_version_api,
    pulp_requests,
):
    """A fixture that fetches the content from a repository."""

    def _get_content(repo, version_href=None):
        """Fetches the content from a given repository.

        Args:
            repo: The repository where the content is fetched from.
            version_href: The repository version from where the content should be fetched from.
                Default: latest repository version.

        Returns:
            A dictionary with lists of packages by content_type (package, modulemd, etc)
            for 'present', 'added' and 'removed' content. E.g:

            ```python
            >>> get_content(repository)
            {
                'present': {
                    'rpm.package': [{'arch', 'noarch', 'artifact': ...}],
                    'rpm.packagegroup': [{'arch', 'noarch', 'artifact': ...}],
                    ...
                },
                'added': { ... },
                'removed': { ... },
            }
            ```
        """

        def fetch_content(pulp_href) -> list:
            result = pulp_requests.get(pulp_href)
            result.raise_for_status()
            return result.json()["results"]

        # Select verion_href
        version_href = version_href or repo.latest_version_href
        if version_href is None:
            return {}
        content_summary = rpm_repository_version_api.read(version_href).content_summary

        result = {}
        for key in ("present", "added", "removed"):
            content = {}
            # ensure every content type returns at least an empty list
            for k in RPM_CONTENT_NAMES:
                content[k] = []
            # fetch content details for each content type
            summary_entry = getattr(content_summary, key)
            for content_type, content_dict in summary_entry.items():
                content[content_type] = fetch_content(content_dict["href"])
            result[key] = content
        return result

    return _get_content



from .acs import RpmAlternateContentSourceViewSet  # noqa
from .advisory import UpdateRecordViewSet  # noqa
from .comps import (  # noqa
    CompsXmlViewSet,
    PackageGroupViewSet,
    PackageCategoryViewSet,
    PackageEnvironmentViewSet,
    PackageLangpacksViewSet,
)
from .custom_metadata import RepoMetadataFileViewSet  # noqa
from .distribution import DistributionTreeViewSet  # noqa
from .modulemd import ModulemdViewSet, ModulemdDefaultsViewSet, ModulemdObsoleteViewSet  # noqa
from .package import PackageViewSet  # noqa
from .prune import PrunePackagesViewSet  # noqa
from .repository import (  # noqa
    RpmRepositoryViewSet,
    RpmRepositoryVersionViewSet,
    RpmRemoteViewSet,
    UlnRemoteViewSet,
    RpmPublicationViewSet,
    RpmDistributionViewSet,
    CopyViewSet,
)



# Generated by Django 4.2.10 on 2024-04-15 18:44

import hashlib

from django.db import migrations


def add_snippet_hash(apps, schema_editor):
    """Calculate and add digest hash of the snippet."""

    ModulemdDefaults = apps.get_model("rpm", "ModulemdDefaults")
    modules_to_update = []
    for mmd in ModulemdDefaults.objects.filter(digest__in=('', None)).only("snippet").iterator():
        mmd.digest = hashlib.sha256(mmd.snippet.encode()).hexdigest()
        modules_to_update.append(mmd)
    ModulemdDefaults.objects.bulk_update(modules_to_update, fields=["digest"])

    Modulemd = apps.get_model("rpm", "Modulemd")
    modules_to_update = []
    for mmd in Modulemd.objects.filter(digest__in=('', None)).only("snippet").iterator():
        mmd.digest = hashlib.sha256(mmd.snippet.encode()).hexdigest()
        modules_to_update.append(mmd)
    Modulemd.objects.bulk_update(modules_to_update, fields=["digest"])


class Migration(migrations.Migration):
    dependencies = [
        ("rpm", "0060_rpmpublication_compression_type_empty"),
    ]

    operations = [
        migrations.RunPython(
            code=add_snippet_hash,
            reverse_code=migrations.RunPython.noop,
        ),
    ]



from datetime import datetime, timedelta
from logging import getLogger, DEBUG

from django.conf import settings
from django.db.models import F, Subquery
from django.utils import timezone

from pulpcore.plugin.models import ProgressReport
from pulpcore.plugin.constants import TASK_STATES
from pulpcore.plugin.models import (
    GroupProgressReport,
    RepositoryContent,
    TaskGroup,
)
from pulpcore.plugin.tasking import dispatch
from pulp_rpm.app.models.package import Package
from pulp_rpm.app.models.repository import RpmRepository

log = getLogger(__name__)


def prune_repo_packages(repo_pk, keep_days, dry_run):
    """
    This task prunes old Packages from the latest_version of the specified repository.

    Args:
        repo_pk (UUID): UUID of the RpmRepository to be pruned.
        keep_days(int): Keep RepositoryContent created less than this many days ago.
        dry_run (boolean): If True, don't actually do the prune, just log to-be-pruned Packages.
    """
    repo = RpmRepository.objects.filter(pk=repo_pk).get()
    curr_vers = repo.latest_version()
    eldest_datetime = datetime.now(tz=timezone.utc) - timedelta(days=keep_days)
    log.info(f"PRUNING REPOSITORY {repo.name}.")
    log.debug(f">>> TOTAL RPMS: {curr_vers.get_content(Package.objects).count()}")

    # We only care about RPM-Names that have more than one EVRA - "singles" are always kept.
    rpm_by_name_age = (
        curr_vers.get_content(Package.objects.with_age())
        .filter(age__gt=1)
        .order_by("name", "epoch", "version", "release", "arch")
        .values("pk")
    )
    log.debug(f">>> NAME/AGE COUNT {rpm_by_name_age.count()}")
    log.debug(
        ">>> # NAME/ARCH w/ MULTIPLE EVRs: {}".format(
            curr_vers.get_content(Package.objects)
            .filter(pk__in=rpm_by_name_age)
            .values("name", "arch")
            .distinct()
            .count()
        )
    )
    log.debug(
        ">>> # UNIQUE NAME/ARCHS: {}".format(
            curr_vers.get_content(Package.objects).values("name", "arch").distinct().count()
        )
    )

    # Find the RepositoryContents associated with the multi-EVR-names from above,
    # whose maximum-pulp-created date is LESS THAN eldest_datetime.
    #
    # Note that we can "assume" the latest-date is an "add" with no "remove", since we're
    # limiting ourselves to the list of ids that we know are in the repo's current latest-version!
    target_ids_q = (
        RepositoryContent.objects.filter(
            content__in=Subquery(rpm_by_name_age), repository=repo, version_removed=None
        )
        .filter(pulp_created__lt=eldest_datetime)
        .values("content_id")
    )
    log.debug(f">>> TARGET IDS: {target_ids_q.count()}.")
    to_be_removed = target_ids_q.count()
    # Use the progressreport to report back numbers. The prune happens as one
    # action.
    data = dict(
        message=f"Pruning {repo.name}",
        code="rpm.package.prune.repository",
        total=to_be_removed,
        state=TASK_STATES.COMPLETED,
        done=0,
    )

    if dry_run:
        if log.getEffectiveLevel() == DEBUG:  # Don't go through the loop unless debugging
            log.debug(">>> Packages to be removed : ")
            for p in (
                Package.objects.filter(pk__in=target_ids_q)
                .order_by("name", "epoch", "version", "release", "arch")
                .values("name", "epoch", "version", "release", "arch")
            ):
                log.debug(f'{p["name"]}-{p["epoch"]}:{p["version"]}-{p["release"]}.{p["arch"]}')
    else:
        with repo.new_version(base_version=None) as new_version:
            new_version.remove_content(target_ids_q)
        data["done"] = to_be_removed

    pb = ProgressReport(**data)
    pb.save()

    # Report back that this repo has completed.
    gpr = TaskGroup.current().group_progress_reports.filter(code="rpm.package.prune")
    gpr.update(done=F("done") + 1)


def prune_packages(
    repo_pks,
    keep_days=14,
    dry_run=False,
):
    """
    This task prunes old Packages from the latest_version of the specified list of repos.

    "Old" in this context is defined by the RepositoryContent record that added a Package
    to the repository in question.

    It will issue one task-per-repository.

    Kwargs:
        repo_pks (list): A list of repo pks the pruning is performed on.
        keep_days(int): Keep RepositoryContent created less than this many days ago.
        repo_concurrency (int): number of repos to prune at a time.
        dry_run (boolean): If True, don't actually do the prune, just record to-be-pruned Packages..
    """

    repos_to_prune = RpmRepository.objects.filter(pk__in=repo_pks)
    task_group = TaskGroup.current()

    # We want to be able to limit the number of available-workers that prune will consume,
    # so that pulp can continue to work while pruning many repositories. We accomplish this by
    # creating a reserved-resource string for each repo-prune-task based on that repo's index in
    # the dispatch loop, mod number-of-workers-to-consume.
    #
    # By default, prune will consume up to 5 workers.
    #
    # (This comment and code below based on
    #   https://github.com/pulp/pulpcore/blob/main/pulpcore/app/tasks/importer.py#L503-L512
    # When we have a generic-approach to throttling mass-task-spawning, both places should
    # be refactored to take advantage thereof.
    prune_workers = int(settings.get("PRUNE_WORKERS_MAX", 5))

    gpr = GroupProgressReport(
        message="Pruning old Packages",
        code="rpm.package.prune",
        total=len(repo_pks),
        done=0,
        task_group=task_group,
    )
    gpr.save()

    # Dispatch a task-per-repository.
    # Lock on the the repository *and* to insure the max-concurrency specified.
    # This will keep an "all repositories" prune from locking up all the workers
    # until all repositories are completed.
    for index, a_repo in enumerate(repos_to_prune):
        worker_rsrc = f"rpm-prune-worker-{index % prune_workers}"
        exclusive_resources = [worker_rsrc, a_repo]

        dispatch(
            prune_repo_packages,
            exclusive_resources=exclusive_resources,
            args=(
                a_repo.pk,
                keep_days,
                dry_run,
            ),
            task_group=task_group,
        )



import json

import pytest
import uuid

from django.conf import settings

from pulpcore.client.pulp_rpm import ApiException, Copy, RpmRepositorySyncURL
from pulpcore.client.pulpcore.exceptions import ApiException as CoreApiException

from pulp_rpm.tests.functional.utils import (
    get_package_repo_path,
)
from pulp_rpm.tests.functional.constants import (
    RPM_SIGNED_FIXTURE_URL,
)

if not settings.DOMAIN_ENABLED:
    pytest.skip("Domains not enabled.", allow_module_level=True)


def test_domain_create(
    pulpcore_bindings,
    gen_object_with_cleanup,
    monitor_task,
    rpm_package_api,
    rpm_repository_api,
    rpm_rpmremote_api,
    rpm_rpmremote_factory,
):
    """Test repo-creation in a domain."""
    body = {
        "name": str(uuid.uuid4()),
        "storage_class": "pulpcore.app.models.storage.FileSystem",
        "storage_settings": {"MEDIA_ROOT": "/var/lib/pulp/media/"},
    }
    domain = gen_object_with_cleanup(pulpcore_bindings.DomainsApi, body)
    domain_name = domain.name

    # create and sync in default domain (not specified)
    remote = rpm_rpmremote_factory(url=RPM_SIGNED_FIXTURE_URL)
    repo_body = {"name": str(uuid.uuid4()), "remote": remote.pulp_href}
    repo = gen_object_with_cleanup(rpm_repository_api, repo_body)
    # Check that we can "find" the new repo in the default-domain
    repo = rpm_repository_api.read(repo.pulp_href)
    sync_url = RpmRepositorySyncURL(remote=remote.pulp_href)
    monitor_task(rpm_repository_api.sync(repo.pulp_href, sync_url).task)

    # check that newly created domain doesn't have a repo or any packages
    assert rpm_repository_api.list(pulp_domain=domain_name).count == 0
    assert rpm_package_api.list(pulp_domain=domain_name).count == 0


def test_domain_sync(
    cleanup_domains,
    pulpcore_bindings,
    gen_object_with_cleanup,
    monitor_task,
    rpm_advisory_api,
    rpm_package_api,
    rpm_package_category_api,
    rpm_package_groups_api,
    rpm_package_lang_packs_api,
    rpm_repository_api,
    rpm_rpmremote_api,
    rpm_rpmremote_factory,
):
    """Test repo-sync in a domain."""

    try:
        body = {
            "name": str(uuid.uuid4()),
            "storage_class": "pulpcore.app.models.storage.FileSystem",
            "storage_settings": {"MEDIA_ROOT": "/var/lib/pulp/media/"},
        }
        domain = gen_object_with_cleanup(pulpcore_bindings.DomainsApi, body)
        domain_name = domain.name

        # create and sync in the newly-created domain
        remote = rpm_rpmremote_factory(
            name=str(uuid.uuid4()), url=RPM_SIGNED_FIXTURE_URL, pulp_domain=domain_name
        )
        repo_body = {"name": str(uuid.uuid4()), "remote": remote.pulp_href}
        repo = gen_object_with_cleanup(rpm_repository_api, repo_body, pulp_domain=domain_name)

        # Check that we can "find" the new repo in the new domain via filtering
        repos = rpm_repository_api.list(name=repo.name, pulp_domain=domain_name).results
        assert len(repos) == 1
        assert repos[0].pulp_href == repo.pulp_href
        sync = RpmRepositorySyncURL()
        monitor_task(rpm_repository_api.sync(repo.pulp_href, sync).task)
        repo = rpm_repository_api.read(repo.pulp_href)

        # check that newly created domain has one repo (list works) and the expected contents
        assert rpm_repository_api.list(pulp_domain=domain_name).count == 1
        assert (
            rpm_package_api.list(
                repository_version=repo.latest_version_href, pulp_domain=domain_name
            ).count
            == 35
        )
        assert (
            rpm_advisory_api.list(
                repository_version=repo.latest_version_href, pulp_domain=domain_name
            ).count
            == 4
        )

        assert (
            rpm_package_category_api.list(
                repository_version=repo.latest_version_href, pulp_domain=domain_name
            ).count
            == 1
        )
        assert (
            rpm_package_groups_api.list(
                repository_version=repo.latest_version_href, pulp_domain=domain_name
            ).count
            == 2
        )
        assert (
            rpm_package_lang_packs_api.list(
                repository_version=repo.latest_version_href, pulp_domain=domain_name
            ).count
            == 1
        )
    finally:
        cleanup_domains([domain], content_api_client=rpm_package_api, cleanup_repositories=True)


@pytest.mark.parallel
def test_object_creation(
    pulpcore_bindings,
    gen_object_with_cleanup,
    rpm_repository_api,
    rpm_rpmremote_api,
    rpm_rpmremote_factory,
):
    """Test basic object creation in a separate domain."""
    body = {
        "name": str(uuid.uuid4()),
        "storage_class": "pulpcore.app.models.storage.FileSystem",
        "storage_settings": {"MEDIA_ROOT": "/var/lib/pulp/media/"},
    }
    domain = gen_object_with_cleanup(pulpcore_bindings.DomainsApi, body)
    domain_name = domain.name

    repo_body = {"name": str(uuid.uuid4())}
    repo = gen_object_with_cleanup(rpm_repository_api, repo_body, pulp_domain=domain_name)
    assert f"{domain_name}/api/v3/" in repo.pulp_href

    repos = rpm_repository_api.list(pulp_domain=domain_name)
    assert repos.count == 1
    assert repo.pulp_href == repos.results[0].pulp_href

    # list repos on default domain
    default_repos = rpm_repository_api.list(name=repo.name)
    assert default_repos.count == 0

    # Try to create an object w/ cross domain relations
    default_remote = rpm_rpmremote_factory()
    with pytest.raises(ApiException) as e:
        repo_body = {"name": str(uuid.uuid4()), "remote": default_remote.pulp_href}
        rpm_repository_api.create(repo_body, pulp_domain=domain.name)
    assert e.value.status == 400
    # What key should this error be under? non-field-errors seems wrong
    assert json.loads(e.value.body) == {
        "non_field_errors": [f"Objects must all be a part of the {domain_name} domain."]
    }

    with pytest.raises(ApiException) as e:
        sync_body = {"remote": default_remote.pulp_href}
        rpm_repository_api.sync(repo.pulp_href, sync_body)
    assert e.value.status == 400
    assert json.loads(e.value.body) == {
        "non_field_errors": [f"Objects must all be a part of the {domain_name} domain."]
    }


@pytest.mark.parallel
def test_artifact_from_file(
    pulpcore_bindings,
    gen_object_with_cleanup,
    rpm_artifact_factory,
):
    """Test uploading artifacts in separate domains."""
    body = {
        "name": str(uuid.uuid4()),
        "storage_class": "pulpcore.app.models.storage.FileSystem",
        "storage_settings": {"MEDIA_ROOT": "/var/lib/pulp/media/"},
    }
    domain1 = gen_object_with_cleanup(pulpcore_bindings.DomainsApi, body)

    body = {
        "name": str(uuid.uuid4()),
        "storage_class": "pulpcore.app.models.storage.FileSystem",
        "storage_settings": {"MEDIA_ROOT": "/var/lib/pulp/media/"},
    }
    domain2 = gen_object_with_cleanup(pulpcore_bindings.DomainsApi, body)

    # Create as-artifact in domain1
    domain1_artifact = rpm_artifact_factory(pulp_domain=domain1.name)
    artifacts = pulpcore_bindings.ArtifactsApi.list(pulp_domain=domain1.name)
    assert artifacts.count == 1
    assert domain1_artifact.pulp_href == artifacts.results[0].pulp_href

    # Create as-artifact in domain2
    domain2_artifact = rpm_artifact_factory(pulp_domain=domain2.name)
    artifacts = pulpcore_bindings.ArtifactsApi.list(pulp_domain=domain2.name)
    assert artifacts.count == 1
    assert domain2_artifact.pulp_href == artifacts.results[0].pulp_href

    # Show same artifact, diff domains
    assert domain1_artifact.pulp_href != domain2_artifact.pulp_href
    assert domain1_artifact.sha256 == domain2_artifact.sha256

    # Show that duplicate artifact can not be uploaded in same domain
    with pytest.raises(CoreApiException) as e:
        rpm_artifact_factory(pulp_domain=domain1.name)
    assert e.value.status == 400
    assert json.loads(e.value.body) == {
        "non_field_errors": [
            f"Artifact with sha256 checksum of '{domain1_artifact.sha256}' already exists.",
        ]
    }


@pytest.mark.parallel
def test_rpm_from_file(
    cleanup_domains,
    pulpcore_bindings,
    rpm_package_factory,
    gen_object_with_cleanup,
    rpm_package_api,
):
    """Test uploading of RPM content with domains."""
    body = {
        "name": str(uuid.uuid4()),
        "storage_class": "pulpcore.app.models.storage.FileSystem",
        "storage_settings": {"MEDIA_ROOT": "/var/lib/pulp/media/"},
    }
    domain = gen_object_with_cleanup(pulpcore_bindings.DomainsApi, body)

    try:
        default_content = rpm_package_factory()
        domain_content = rpm_package_factory(pulp_domain=domain.name)
        assert default_content.pulp_href != domain_content.pulp_href
        assert default_content.sha256 == domain_content.sha256

        domain_contents = rpm_package_api.list(pulp_domain=domain.name)
        assert domain_contents.count == 1
    finally:
        cleanup_domains([domain], content_api_client=rpm_package_api)


@pytest.mark.parallel
def test_content_promotion(
    cleanup_domains,
    distribution_base_url,
    pulpcore_bindings,
    download_content_unit,
    rpm_repository_api,
    rpm_rpmremote_factory,
    rpm_publication_api,
    rpm_distribution_factory,
    gen_object_with_cleanup,
    monitor_task,
):
    """Tests Content promotion path with domains: Sync->Publish->Distribute"""
    body = {
        "name": str(uuid.uuid4()),
        "storage_class": "pulpcore.app.models.storage.FileSystem",
        "storage_settings": {"MEDIA_ROOT": "/var/lib/pulp/media/"},
    }
    domain = gen_object_with_cleanup(pulpcore_bindings.DomainsApi, body)

    try:
        # Sync task
        remote = rpm_rpmremote_factory(pulp_domain=domain.name)
        repo_body = {"name": str(uuid.uuid4()), "remote": remote.pulp_href}
        repo = rpm_repository_api.create(repo_body, pulp_domain=domain.name)

        task = rpm_repository_api.sync(repo.pulp_href, {}).task
        response = monitor_task(task)
        assert len(response.created_resources) == 1

        repo = rpm_repository_api.read(repo.pulp_href)
        assert repo.latest_version_href[-2] == "1"

        # Publish task
        pub_body = {"repository": repo.pulp_href}
        task = rpm_publication_api.create(pub_body, pulp_domain=domain.name).task
        response = monitor_task(task)
        assert len(response.created_resources) == 1
        pub_href = response.created_resources[0]
        pub = rpm_publication_api.read(pub_href)

        assert pub.repository == repo.pulp_href

        # Distribute Task
        distro = rpm_distribution_factory(publication=pub.pulp_href, pulp_domain=domain.name)

        assert distro.publication == pub.pulp_href
        # Url structure should be host/CONTENT_ORIGIN/DOMAIN_PATH/BASE_PATH
        assert domain.name == distribution_base_url(distro.base_url).rstrip("/").split("/")[-2]

        # Check that content can be downloaded from base_url
        for pkg in ("bear-4.1-1.noarch.rpm", "pike-2.2-1.noarch.rpm"):
            pkg_path = get_package_repo_path(pkg)
            download_content_unit(distro.base_path, pkg_path, domain=domain.name)

        # Cleanup to delete the domain
        task = rpm_repository_api.delete(repo.pulp_href).task
        monitor_task(task)
    finally:
        cleanup_domains([domain], cleanup_repositories=True)


@pytest.mark.parallel
def test_domain_rbac(
    cleanup_domains, pulpcore_bindings, gen_user, gen_object_with_cleanup, rpm_repository_api
):
    """Test domain level-roles."""
    body = {
        "name": str(uuid.uuid4()),
        "storage_class": "pulpcore.app.models.storage.FileSystem",
        "storage_settings": {"MEDIA_ROOT": "/var/lib/pulp/media/"},
    }
    domain = gen_object_with_cleanup(pulpcore_bindings.DomainsApi, body)

    try:
        rpm_viewer = "rpm.rpmrepository_viewer"
        rpm_creator = "rpm.rpmrepository_creator"
        user_a = gen_user(username="a", domain_roles=[(rpm_viewer, domain.pulp_href)])
        user_b = gen_user(username="b", domain_roles=[(rpm_creator, domain.pulp_href)])

        # Create two repos in different domains w/ admin user
        gen_object_with_cleanup(rpm_repository_api, {"name": str(uuid.uuid4())})
        gen_object_with_cleanup(
            rpm_repository_api, {"name": str(uuid.uuid4())}, pulp_domain=domain.name
        )

        with user_b:
            repo = gen_object_with_cleanup(
                rpm_repository_api, {"name": str(uuid.uuid4())}, pulp_domain=domain.name
            )
            repos = rpm_repository_api.list(pulp_domain=domain.name)
            assert repos.count == 1
            assert repos.results[0].pulp_href == repo.pulp_href
            # Try to create a repository in default domain
            with pytest.raises(ApiException) as e:
                rpm_repository_api.create({"name": str(uuid.uuid4())})
            assert e.value.status == 403

        with user_a:
            repos = rpm_repository_api.list(pulp_domain=domain.name)
            assert repos.count == 2
            # Try to read repos in the default domain
            repos = rpm_repository_api.list()
            assert repos.count == 0
            # Try to create a repo
            with pytest.raises(ApiException) as e:
                rpm_repository_api.create({"name": str(uuid.uuid4())}, pulp_domain=domain.name)
            assert e.value.status == 403

    finally:
        cleanup_domains([domain], cleanup_repositories=True)


@pytest.mark.parallel
def test_advisory_upload_json(
    cleanup_domains,
    setup_domain,
    upload_advisory_factory,
):
    """Test upload same advisory from JSON file into different Domains."""
    domain1 = None
    domain2 = None
    try:
        domain1, _, repo1, _ = setup_domain(sync=False)
        domain2, _, repo2, _ = setup_domain(sync=False)

        advisory1, _, an_id = upload_advisory_factory(
            pulp_domain=domain1.name, repository=repo1, set_id=True
        )
        assert advisory1.id == an_id

        advisory2, _, an_id = upload_advisory_factory(
            pulp_domain=domain2.name, repository=repo2, set_id=True
        )
        assert advisory2.id == an_id

        assert advisory1.pulp_href != advisory2.pulp_href
    finally:
        cleanup_domains([domain1, domain2], cleanup_repositories=True)


@pytest.mark.parallel
def test_cross_domain_copy_all(
    monitor_task,
    rpm_copy_api,
    cleanup_domains,
    setup_domain,
):
    """Test attempting to copy between different Domains."""
    domain1 = None
    domain2 = None
    try:
        domain1, remote1, src1, dest1 = setup_domain()
        domain2, remote2, src2, dest2 = setup_domain()
        # Success, everything in domain1
        data = Copy(
            config=[
                {"source_repo_version": src1.latest_version_href, "dest_repo": dest1.pulp_href}
            ],
            dependency_solving=False,
        )
        monitor_task(rpm_copy_api.copy_content(data, pulp_domain=domain1.name).task)

        # Failure, call and src domain1, dest domain2
        with pytest.raises(ApiException):
            data = Copy(
                config=[
                    {"source_repo_version": src1.latest_version_href, "dest_repo": dest2.pulp_href}
                ],
                dependency_solving=False,
            )
            monitor_task(rpm_copy_api.copy_content(data, pulp_domain=domain1.name).task)

        # Failure, call domain2, src/dest domain1
        with pytest.raises(ApiException):
            data = Copy(
                config=[
                    {"source_repo_version": src1.latest_version_href, "dest_repo": dest1.pulp_href}
                ],
                dependency_solving=False,
            )
            monitor_task(rpm_copy_api.copy_content(data, pulp_domain=domain2.name).task)

    finally:
        cleanup_domains([domain1, domain2], cleanup_repositories=True)


@pytest.mark.parallel
def test_cross_domain_content(
    cleanup_domains,
    monitor_task,
    rpm_advisory_api,
    rpm_copy_api,
    rpm_repository_api,
    setup_domain,
):
    """Test the content parameter."""
    domain1 = None
    domain2 = None
    try:
        domain1, remote1, src1, dest1 = setup_domain()
        domain2, remote2, src2, dest2 = setup_domain()

        # Copy content1 from src1 to dest1, expect 2 copied advisories
        advisories1 = rpm_advisory_api.list(
            repository_version=src1.latest_version_href, pulp_domain=domain1.name
        ).results
        advisories_to_copy1 = (advisories1[0].pulp_href, advisories1[1].pulp_href)

        data = Copy(
            config=[
                {
                    "source_repo_version": src1.latest_version_href,
                    "dest_repo": dest1.pulp_href,
                    "content": advisories_to_copy1,
                }
            ],
            dependency_solving=False,
        )
        monitor_task(rpm_copy_api.copy_content(data, pulp_domain=domain1.name).task)
        dest1 = rpm_repository_api.read(dest1.pulp_href)
        advisories = rpm_advisory_api.list(
            repository_version=dest1.latest_version_href, pulp_domain=domain1.name
        ).results
        assert 2 == len(advisories)

        # Copy content1 from src1 to dest1, domain2, expect failure
        with pytest.raises(ApiException):
            data = Copy(
                config=[
                    {
                        "source_repo_version": src1.latest_version_href,
                        "dest_repo": dest2.pulp_href,
                        "content": advisories_to_copy1,
                    }
                ],
                dependency_solving=False,
            )
            monitor_task(rpm_copy_api.copy_content(data, pulp_domain=domain1.name).task)

        # Copy content1 from src1 to dest2, domain1, expect failure
        with pytest.raises(ApiException):
            data = Copy(
                config=[
                    {
                        "source_repo_version": src1.latest_version_href,
                        "dest_repo": dest2.pulp_href,
                        "content": advisories_to_copy1,
                    }
                ],
                dependency_solving=False,
            )
            monitor_task(rpm_copy_api.copy_content(data, pulp_domain=domain1.name).task)

        # Copy mixed content from src2 to dest2, domain2, expect failure
        advisories2 = rpm_advisory_api.list(
            repository_version=src2.latest_version_href, pulp_domain=domain2.name
        ).results
        advisories_to_copy2 = (advisories2[0].pulp_href, advisories2[1].pulp_href)

        with pytest.raises(ApiException):
            data = Copy(
                config=[
                    {
                        "source_repo_version": src2.latest_version_href,
                        "dest_repo": dest2.pulp_href,
                        "content": advisories_to_copy1 + advisories_to_copy2,
                    }
                ],
                dependency_solving=False,
            )
            task = rpm_copy_api.copy_content(data, pulp_domain=domain2.name).task
            monitor_task(task)

    finally:
        cleanup_domains([domain1, domain2], cleanup_repositories=True)



import os
import asyncio

from aiohttp_xmlrpc.client import ServerProxy, _Method
from logging import getLogger
from lxml import etree
from urllib.parse import quote, unquote, urlparse

from pulpcore.plugin.download import FileDownloader, HttpDownloader
from pulp_rpm.app.exceptions import UlnCredentialsError
from pulp_rpm.app.shared_utils import urlpath_sanitize


log = getLogger(__name__)


class RpmFileDownloader(FileDownloader):
    """
    FileDownloader that strips out RPM's custom http downloader arguments.

    This is unfortunate, but there isn't currently a good pattern for customizing the downloader
    factory machinery such that certain types of arguments only apply to certain downloaders,
    so passing a kwarg into get_downloader() will pass it to constructor for any downloader.

    TODO: https://pulp.plan.io/issues/7352
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the downloader.
        """
        kwargs.pop("silence_errors_for_response_status_codes", None)
        super().__init__(*args, **kwargs)


class RpmDownloader(HttpDownloader):
    """
    Custom Downloader that automatically handles authentication token for SLES repositories.

    Args:
        silence_errors_for_response_status_codes (iterable): An iterable of response exception
            codes to be ignored when raising exception. e.g. `{404}`
        sles_auth_token (str): SLES authentication token.

    Raises:
        FileNotFoundError: If aiohttp response status is 404 and silenced.
    """

    def __init__(
        self,
        *args,
        silence_errors_for_response_status_codes=None,
        sles_auth_token=None,
        urlencode=True,
        **kwargs,
    ):
        """
        Initialize the downloader.
        """
        self.sles_auth_token = sles_auth_token

        if silence_errors_for_response_status_codes is None:
            silence_errors_for_response_status_codes = set()
        self.silence_errors_for_response_status_codes = silence_errors_for_response_status_codes

        super().__init__(*args, **kwargs)

        new_url = self.url
        if urlencode:
            # Some upstream-repos (eg, Amazon) require url-encoded paths for things like "libc++"
            # Let's make them happy.
            # We can't urlencode the whole url, because BasicAuth is still A Thing and we would
            #   break username/passwords in the url.
            # So we need to urlencode **only the path** and **nothing else** .
            # We can't use _replace() and urlunparse(), because urlunparse() "helpfully" undoes
            #   the urlencode we just did in the path.
            # We can't use urljoin(), because urljoin() "helpfully" treats a number of schemes
            #  (like, say, uln:) as "can't take relative paths", and throws away everything
            #  **except** the path-portion
            # So, we have a pretty ugly workaround.
            parsed = urlparse(self.url)
            # two pieces of the URL: pre- and post-path
            (before_path, after_path) = self.url.split(parsed.path)
            new_path = quote(unquote(parsed.path), safe=":/")  # fix the path
            new_url = "{}{}{}".format(before_path, new_path, after_path)  # rebuild
        if self.sles_auth_token:
            auth_param = f"?{self.sles_auth_token}"
            self.url = urlpath_sanitize(new_url) + auth_param
        else:
            self.url = new_url

    def raise_for_status(self, response):
        """
        Raise error if aiohttp response status is >= 400 and not silenced.

        Raises:
            FileNotFoundError: If aiohttp response status is 403 or 404 and silenced.
            aiohttp.ClientResponseError: If the response status is 400 or higher and not silenced.
        """
        silenced = response.status in self.silence_errors_for_response_status_codes

        if not silenced:
            response.raise_for_status()

        if response.status in (404, 403):
            raise FileNotFoundError()

    async def _run(self, extra_data=None):
        """
        Download, validate, and compute digests on the `url`. This is a coroutine.

        This method provides the same return object type and documented in
        :meth:`~pulpcore.plugin.download.BaseDownloader._run`.
        """
        async with self.session.get(
            self.url, proxy=self.proxy, proxy_auth=self.proxy_auth, auth=self.auth
        ) as response:
            self.raise_for_status(response)
            to_return = await self._handle_response(response)
            await response.release()
            self.response_headers = response.headers

        if self._close_session_on_finalize:
            self.session.close()
        return to_return


class UlnDownloader(RpmDownloader):
    """
    Custom Downloader for ULN repositories.

    Args:
        username (str): Username for authentication in ULN network
        password (str): password for authentication in ULN network
        uln_server_base_url (str): ULN server url.

    Raises:
        UlnCredentialsError: If no or not valid ULN credentials are given,
            this Error will be displayed
    """

    def __init__(self, *args, username=None, password=None, uln_server_base_url=None, **kwargs):
        """
        Initialize the downloader for ULN repositories.

        If no server URL is given, use the server url from oracle linux.
        """
        self.username = username
        self.password = password
        self.uln_server_base_url = uln_server_base_url
        self.headers = None
        self.session_key = None

        super().__init__(*args, **kwargs)

    async def _run(self, extra_data=None):
        """
        Download, validate, and compute digests on the `url`. This is a coroutine.

        Once per session the coroutine logs into the ULN account using the
        ULN username and password. The returned key is used for authentification
        for all other downloads.

        This method provides the same return object type and documented in
        :meth:`~pulpcore.plugin.download.BaseDownloader._run`.
        """
        parsed = urlparse(self.url)

        if parsed.scheme == "uln":
            # get ULN Session-key
            SERVER_URL = os.path.join(self.uln_server_base_url, "rpc/api")

            # set proxy for authentification
            client = AllowProxyServerProxy(
                SERVER_URL, proxy=self.proxy, proxy_auth=self.proxy_auth, auth=self.auth
            )
            if not self.session_key:
                self.session_key = await self._login_and_retry(client)
                if len(self.session_key) != 43:
                    raise UlnCredentialsError()
                self.headers = {"X-ULN-API-User-Key": self.session_key}
                await client.close()
            # build request url from input uri
            channelLabel = parsed.netloc
            path = parsed.path.lstrip("/")
            url = os.path.join(self.uln_server_base_url, "XMLRPC/GET-REQ", channelLabel, path)
        async with self.session.get(
            url, proxy=self.proxy, proxy_auth=self.proxy_auth, auth=self.auth, headers=self.headers
        ) as response:
            self.raise_for_status(response)
            to_return = await self._handle_response(response)
            await response.release()
            self.response_headers = response.headers

        if self._close_session_on_finalize:
            self.session.close()
            client.close()
        return to_return

    async def _login_and_retry(self, client, max_attempts=4, delay=1):
        """
        Attempts to log in using the provided client. Retries up to `max_attempts`
        times with a delay between attempts.

        Args:
            client: The client object used for authentication.
            max_attempts (int): Maximum number of login attempts.
            delay (int or float): Delay in seconds between retries.

        Returns:
            The session key from the login.

        Raises:
            UlnCredentialsError: If all login attempts fail.
        """
        for attempt in range(1, max_attempts + 1):
            try:
                return await client.auth.login(self.username, self.password)
            except Exception as exc:
                if attempt < max_attempts:
                    log.info(f"ULN login attempt {attempt} failed. Retrying...")
                    await asyncio.sleep(delay)
                else:
                    raise UlnCredentialsError(
                        f"ULN login failed after {max_attempts} attempts."
                    ) from exc


class AllowProxyServerProxy(ServerProxy):
    """
    Overwriting the class aiohttp_xmlrpc.ServreProxy to allow proxy handling.

    Until aiohttp-xmlrpc allows http post with proxy, use this patch.
    This only works for http connection to the proxy, https connections are not supported!
    """

    def __init__(self, *args, proxy, proxy_auth, auth, **kwargs):
        """
        Initialisation with proxy.
        """
        self.__proxy = proxy
        self.__proxy_auth = proxy_auth
        self.__auth = auth
        super().__init__(*args, **kwargs)

    async def __remote_call(self, method_name, *args, **kwargs):
        """
        Set proxy for HTTP POST call.
        """
        async with self.client.post(
            str(self.url),
            data=etree.tostring(
                self._make_request(method_name, *args, **kwargs),
                xml_declaration=True,
                encoding=self.encoding,
            ),
            headers=self.headers,
            proxy=self.__proxy,
            proxy_auth=self.__proxy_auth,
            auth=self.__auth,
        ) as response:
            response.raise_for_status()

            return self._parse_response((await response.read()), method_name)

    # we also have to overwrite this method, because '__'-method-names are only valid
    # in the class they are defined in, therefore only overwriting __remote_call() does not work!
    def __getattr__(self, method_name):
        """
        Dynamically create method for attribute.

        Copied from aiohttp-xmlrpc/aiohttp_xmlrpc/client.py
        """
        # Trick to keep the "close" method available
        if method_name == "close":
            return self.__close
        else:
            # Magic method dispatcher
            return _Method(self.__remote_call, method_name)

    def __close(self):
        return self.client.close()



# Generated by Django 2.2.14 on 2020-08-03 18:58

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0017_merge_advisory_collections'),
    ]

    operations = [
        migrations.AddField(
            model_name='updatecollection',
            name='_update_record',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='_collections', to='rpm.UpdateRecord'),
        ),
    ]



from .publishing import publish  # noqa
from .synchronizing import synchronize  # noqa
from .signing import sign_and_create  # noqa
from .copy import copy_content  # noqa
from .comps import upload_comps  # noqa
from .prune import prune_packages  # noqa



# Generated by Django 4.2.4 on 2023-09-06 14:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("rpm", "0052_modulemd_digest"),
    ]

    operations = [
        migrations.AddField(
            model_name="rpmdistribution",
            name="generate_repo_config",
            field=models.BooleanField(default=False),
        ),
    ]



"""
Check `Plugin Writer's Guide`_ for more details.

.. _Plugin Writer's Guide:
    http://docs.pulpproject.org/plugins/plugin-writer/index.html
"""

DRF_ACCESS_POLICY = {
    "dynaconf_merge_unique": True,
    "reusable_conditions": ["pulp_rpm.app.access_policy"],
}
INSTALLED_APPS = ["django_readonly_field", "dynaconf_merge"]
ALLOW_AUTOMATIC_UNSAFE_ADVISORY_CONFLICT_RESOLUTION = False
DEFAULT_ULN_SERVER_BASE_URL = "https://linux-update.oracle.com/"
KEEP_CHANGELOG_LIMIT = 10
SOLVER_DEBUG_LOGS = True
RPM_METADATA_USE_REPO_PACKAGE_TIME = False
NOCACHE_LIST = ["repomd.xml", "repomd.xml.asc", "repomd.xml.key"]
PRUNE_WORKERS_MAX = 5



from gettext import gettext as _

from rest_framework import serializers

from pulpcore.plugin.serializers import (
    ArtifactSerializer,
    MultipleArtifactContentSerializer,
)

from pulp_rpm.app.models import (
    Addon,
    Checksum,
    DistributionTree,
    Image,
    Variant,
)


class AddonSerializer(serializers.ModelSerializer):
    """
    Addon serializer.
    """

    addon_id = serializers.CharField(help_text=_("Addon id."))
    uid = serializers.CharField(help_text=_("Addon uid."))
    name = serializers.CharField(help_text=_("Addon name."))
    type = serializers.CharField(help_text=_("Addon type."))
    packages = serializers.CharField(help_text=_("Relative path to directory with binary RPMs."))

    class Meta:
        model = Addon
        fields = ("addon_id", "uid", "name", "type", "packages")


class ChecksumSerializer(serializers.ModelSerializer):
    """
    Checksum serializer.
    """

    path = serializers.CharField(help_text=_("File path."))
    checksum = serializers.CharField(help_text=_("Checksum for the file."))

    class Meta:
        model = Checksum
        fields = ("path", "checksum")


class ImageSerializer(serializers.ModelSerializer):
    """
    Image serializer.
    """

    name = serializers.CharField(help_text=_("File name."))
    path = serializers.CharField(help_text=_("File path."))
    platforms = serializers.CharField(help_text=_("Compatible platforms."))
    artifact = ArtifactSerializer(allow_null=True)

    class Meta:
        model = Image
        fields = ("name", "path", "platforms", "artifact")


class VariantSerializer(serializers.ModelSerializer):
    """
    Variant serializer.
    """

    variant_id = serializers.CharField(help_text=_("Variant id."))
    uid = serializers.CharField(help_text=_("Variant uid."))
    name = serializers.CharField(help_text=_("Variant name."))
    type = serializers.CharField(help_text=_("Variant type."))
    packages = serializers.CharField(help_text=_("Relative path to directory with binary RPMs."))
    source_packages = serializers.CharField(
        help_text=_("Relative path to directory with source RPMs."), allow_null=True
    )
    source_repository = serializers.CharField(
        help_text=_("Relative path to YUM repository with source RPMs."), allow_null=True
    )
    debug_packages = serializers.CharField(
        help_text=_("Relative path to directory with debug RPMs."), allow_null=True
    )
    debug_repository = serializers.CharField(
        help_text=_("Relative path to YUM repository with debug RPMs."), allow_null=True
    )
    identity = serializers.CharField(
        help_text=_("Relative path to a pem file that identifies a product."), allow_null=True
    )

    class Meta:
        model = Variant
        fields = (
            "variant_id",
            "uid",
            "name",
            "type",
            "packages",
            "source_packages",
            "source_repository",
            "debug_packages",
            "debug_repository",
            "identity",
        )


class DistributionTreeSerializer(MultipleArtifactContentSerializer):
    """
    DistributionTree serializer.
    """

    header_version = serializers.CharField(help_text=_("Header Version."))
    release_name = serializers.CharField(help_text=_("Release name."))
    release_short = serializers.CharField(help_text=_("Release short name."))
    release_version = serializers.CharField(help_text=_("Release version."))
    release_is_layered = serializers.BooleanField(
        help_text=_("Typically False for an operating system, True otherwise.")
    )

    base_product_name = serializers.CharField(help_text=_("Base Product name."), allow_null=True)
    base_product_short = serializers.CharField(
        help_text=_("Base Product short name."), allow_null=True
    )
    base_product_version = serializers.CharField(
        help_text=_("Base Product version."), allow_null=True
    )

    arch = serializers.CharField(help_text=_("Tree architecturerch."))
    build_timestamp = serializers.FloatField(help_text=_("Tree build time timestamp."))

    instimage = serializers.CharField(
        help_text=_("Relative path to Anaconda instimage."), allow_null=True
    )
    mainimage = serializers.CharField(
        help_text=_("Relative path to Anaconda stage2 image."), allow_null=True
    )

    discnum = serializers.IntegerField(help_text=_("Disc number."), allow_null=True)
    totaldiscs = serializers.IntegerField(
        help_text=_("Number of discs in media set."), allow_null=True
    )

    addons = AddonSerializer(many=True)

    checksums = ChecksumSerializer(many=True)

    images = ImageSerializer(many=True)

    variants = VariantSerializer(many=True)

    class Meta:
        model = DistributionTree
        fields = (
            "pulp_href",
            "prn",
            "header_version",
            "release_name",
            "release_short",
            "release_version",
            "release_is_layered",
            "base_product_name",
            "base_product_short",
            "base_product_version",
            "arch",
            "build_timestamp",
            "instimage",
            "mainimage",
            "discnum",
            "totaldiscs",
            "addons",
            "checksums",
            "images",
            "variants",
        )



import os

from django.utils.timezone import now
from drf_spectacular.utils import extend_schema
from rest_framework.decorators import action
from rest_framework.generics import get_object_or_404

from pulpcore.plugin.models import AlternateContentSourcePath, TaskGroup
from pulpcore.plugin.tasking import dispatch
from pulpcore.plugin.serializers import (
    TaskGroupOperationResponseSerializer,
)
from pulpcore.plugin.viewsets import (
    AlternateContentSourceViewSet,
    RolesMixin,
    TaskGroupOperationResponse,
)

from pulp_rpm.app import tasks
from pulp_rpm.app.constants import SYNC_POLICIES
from pulp_rpm.app.models import (
    RpmAlternateContentSource,
    RpmRepository,
)
from pulp_rpm.app.serializers import (
    RpmAlternateContentSourceSerializer,
    RpmRepositorySyncURLSerializer,
)


class RpmAlternateContentSourceViewSet(AlternateContentSourceViewSet, RolesMixin):
    """
    ViewSet for ACS.
    """

    endpoint_name = "rpm"
    queryset = RpmAlternateContentSource.objects.all()
    serializer_class = RpmAlternateContentSourceSerializer
    queryset_filtering_required_permission = "rpm.view_rpmalternatecontentsource"

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list"],
                "principal": ["authenticated"],
                "effect": "allow",
            },
            {
                "action": ["retrieve"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_domain_or_obj_perms:rpm.view_rpmalternatecontentsource",
            },
            {
                "action": ["create"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_remote_param_model_or_domain_or_obj_perms:rpm.view_rpmremote",
                    "has_model_or_domain_perms:rpm.add_rpmalternatecontentsource",
                ],
            },
            {
                "action": ["refresh"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_model_or_domain_or_obj_perms:rpm.view_rpmalternatecontentsource",
                    "has_model_or_domain_perms:rpm.refresh_rpmalternatecontentsource",
                ],
            },
            {
                "action": ["update", "partial_update"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_model_or_domain_or_obj_perms:rpm.change_rpmalternatecontentsource",
                    "has_model_or_domain_or_obj_perms:rpm.view_rpmalternatecontentsource",
                    "has_remote_param_model_or_domain_or_obj_perms:rpm.view_rpmremote",
                ],
            },
            {
                "action": ["destroy"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_model_or_domain_or_obj_perms:rpm.delete_rpmalternatecontentsource",
                    "has_model_or_domain_or_obj_perms:rpm.view_rpmalternatecontentsource",
                ],
            },
            {
                "action": ["list_roles", "add_role", "remove_role"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_domain_or_obj_perms:"
                "rpm.manage_roles_rpmalternatecontentsource",
            },
        ],
        "queryset_scoping": {"function": "scope_queryset"},
        "creation_hooks": [
            {
                "function": "add_roles_for_object_creator",
                "parameters": {"roles": "rpm.rpmalternatecontentsource_owner"},
            }
        ],
    }

    LOCKED_ROLES = {
        "rpm.rpmalternatecontentsource_owner": [
            "rpm.change_rpmalternatecontentsource",
            "rpm.delete_rpmalternatecontentsource",
            "rpm.manage_roles_rpmalternatecontentsource",
            "rpm.refresh_rpmalternatecontentsource",
            "rpm.view_rpmalternatecontentsource",
        ],
        "rpm.rpmalternatecontentsource_creator": [
            "rpm.add_rpmalternatecontentsource",
        ],
        "rpm.rpmalternatecontentsource_viewer": [
            "rpm.view_rpmalternatecontentsource",
        ],
    }

    @extend_schema(
        description="Trigger an asynchronous task to create Alternate Content Source content.",
        responses={202: TaskGroupOperationResponseSerializer},
        request=None,
    )
    @action(methods=["post"], detail=True)
    def refresh(self, request, pk):
        """
        Refresh ACS metadata.
        """
        acs = get_object_or_404(RpmAlternateContentSource, pk=pk)
        acs_paths = AlternateContentSourcePath.objects.filter(alternate_content_source=pk)
        task_group = TaskGroup.objects.create(
            description=f"Refreshing {acs_paths.count()} alternate content source path(s)."
        )

        # Get required defaults for sync operation
        optimize = RpmRepositorySyncURLSerializer().data["optimize"]
        skip_types = RpmRepositorySyncURLSerializer().data["skip_types"]

        for acs_path in acs_paths:
            # Create or get repository for the path
            repo_data = {
                "name": f"{acs.name}--{acs_path.pk}--repository",
                "retain_repo_versions": 1,
                "user_hidden": True,
            }
            repo, created = RpmRepository.objects.get_or_create(**repo_data)
            if created:
                acs_path.repository = repo
                acs_path.save()
            acs_url = (
                os.path.join(acs.remote.url, acs_path.path) if acs_path.path else acs.remote.url
            )

            # Dispatching ACS path to own task and assign it to common TaskGroup
            dispatch(
                tasks.synchronize,
                shared_resources=[acs.remote, acs],
                task_group=task_group,
                kwargs={
                    "remote_pk": str(acs.remote.pk),
                    "repository_pk": str(acs_path.repository.pk),
                    "sync_policy": SYNC_POLICIES.MIRROR_CONTENT_ONLY,
                    "skip_types": skip_types,
                    "optimize": optimize,
                    "url": acs_url,
                },
            )

        acs.last_refreshed = now()
        acs.save()

        return TaskGroupOperationResponse(task_group, request)



# Generated by Django 4.2.17 on 2024-12-05 18:54

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('core', '0091_systemid'),
    ]

    operations = [
        migrations.CreateModel(
            name='DistributionTree',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='rpm_distributiontree', serialize=False, to='core.content')),
                ('header_version', models.CharField(max_length=10)),
                ('release_name', models.CharField(max_length=50)),
                ('release_short', models.CharField(max_length=20)),
                ('release_version', models.CharField(max_length=10)),
                ('release_is_layered', models.BooleanField(default=False)),
                ('base_product_name', models.CharField(max_length=50, null=True)),
                ('base_product_short', models.CharField(max_length=20, null=True)),
                ('base_product_version', models.CharField(max_length=10, null=True)),
                ('arch', models.CharField(max_length=30)),
                ('build_timestamp', models.FloatField()),
                ('instimage', models.CharField(max_length=50, null=True)),
                ('mainimage', models.CharField(max_length=50, null=True)),
                ('discnum', models.IntegerField(null=True)),
                ('totaldiscs', models.IntegerField(null=True)),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('header_version', 'release_name', 'release_short', 'release_version', 'arch', 'build_timestamp')},
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='ModulemdDefaults',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='rpm_modulemddefaults', serialize=False, to='core.content')),
                ('module', models.CharField(max_length=255)),
                ('stream', models.CharField(max_length=255)),
                ('profiles', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('digest', models.CharField(max_length=64, unique=True)),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='Package',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='rpm_package', serialize=False, to='core.content')),
                ('name', models.CharField(max_length=255)),
                ('epoch', models.CharField(max_length=10)),
                ('version', models.CharField(max_length=255)),
                ('release', models.CharField(max_length=255)),
                ('arch', models.CharField(max_length=20)),
                ('pkgId', models.CharField(max_length=128, unique=True)),
                ('checksum_type', models.CharField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')], max_length=10)),
                ('summary', models.TextField()),
                ('description', models.TextField()),
                ('url', models.TextField()),
                ('changelogs', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('files', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('requires', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('provides', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('conflicts', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('obsoletes', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('suggests', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('enhances', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('recommends', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('supplements', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('location_base', models.TextField()),
                ('location_href', models.TextField()),
                ('rpm_buildhost', models.TextField()),
                ('rpm_group', models.TextField()),
                ('rpm_license', models.TextField()),
                ('rpm_packager', models.TextField()),
                ('rpm_sourcerpm', models.TextField()),
                ('rpm_vendor', models.TextField()),
                ('rpm_header_start', models.BigIntegerField(null=True)),
                ('rpm_header_end', models.BigIntegerField(null=True)),
                ('is_modular', models.BooleanField(default=False)),
                ('size_archive', models.BigIntegerField(null=True)),
                ('size_installed', models.BigIntegerField(null=True)),
                ('size_package', models.BigIntegerField(null=True)),
                ('time_build', models.BigIntegerField(null=True)),
                ('time_file', models.BigIntegerField(null=True)),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('name', 'epoch', 'version', 'release', 'arch', 'checksum_type', 'pkgId')},
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='PackageLangpacks',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='rpm_packagelangpacks', serialize=False, to='core.content')),
                ('matches', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('digest', models.CharField(max_length=64, unique=True)),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='RpmPublication',
            fields=[
                ('publication_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='rpm_rpmpublication', serialize=False, to='core.publication')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.publication',),
        ),
        migrations.CreateModel(
            name='RpmRemote',
            fields=[
                ('remote_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='rpm_rpmremote', serialize=False, to='core.remote')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.remote',),
        ),
        migrations.CreateModel(
            name='RpmRepository',
            fields=[
                ('repository_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='rpm_rpmrepository', serialize=False, to='core.repository')),
                ('sub_repo', models.BooleanField(default=False)),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.repository',),
        ),
        migrations.CreateModel(
            name='UpdateCollection',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('name', models.TextField()),
                ('shortname', models.TextField()),
                ('module', models.TextField(default='')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='UpdateRecord',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='rpm_updaterecord', serialize=False, to='core.content')),
                ('id', models.CharField(db_index=True, max_length=255)),
                ('updated_date', models.TextField()),
                ('description', models.TextField()),
                ('issued_date', models.TextField()),
                ('fromstr', models.TextField()),
                ('status', models.TextField()),
                ('title', models.TextField()),
                ('summary', models.TextField()),
                ('version', models.TextField()),
                ('type', models.TextField()),
                ('severity', models.TextField()),
                ('solution', models.TextField()),
                ('release', models.TextField()),
                ('rights', models.TextField()),
                ('pushcount', models.TextField(blank=True)),
                ('digest', models.CharField(max_length=64, unique=True)),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='UpdateReference',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('href', models.TextField()),
                ('ref_id', models.TextField(null=True)),
                ('title', models.TextField(null=True)),
                ('ref_type', models.TextField()),
                ('update_record', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='references', to='rpm.updaterecord')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='UpdateCollectionPackage',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('arch', models.TextField()),
                ('epoch', models.TextField()),
                ('filename', models.TextField()),
                ('name', models.TextField()),
                ('reboot_suggested', models.BooleanField(default=False)),
                ('release', models.TextField()),
                ('src', models.TextField()),
                ('sum', models.TextField()),
                ('sum_type', models.TextField()),
                ('version', models.TextField()),
                ('update_collection', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='packages', to='rpm.updatecollection')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='updatecollection',
            name='update_record',
            field=models.ManyToManyField(related_name='collections', to='rpm.updaterecord'),
        ),
        migrations.CreateModel(
            name='RpmDistribution',
            fields=[
                ('basedistribution_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='rpm_rpmdistribution', serialize=False, to='core.basedistribution')),
                ('publication', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='rpm_rpmdistribution', to='core.publication')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.basedistribution',),
        ),
        migrations.CreateModel(
            name='RepoMetadataFile',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='rpm_repometadatafile', serialize=False, to='core.content')),
                ('data_type', models.CharField(max_length=20)),
                ('checksum_type', models.CharField(max_length=6)),
                ('checksum', models.CharField(max_length=128)),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('data_type', 'checksum')},
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='PackageGroup',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='rpm_packagegroup', serialize=False, to='core.content')),
                ('id', models.CharField(max_length=255)),
                ('default', models.BooleanField(default=False)),
                ('user_visible', models.BooleanField(default=False)),
                ('display_order', models.IntegerField(null=True)),
                ('name', models.CharField(max_length=255)),
                ('description', models.TextField(default='')),
                ('packages', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('biarch_only', models.BooleanField(default=False)),
                ('desc_by_lang', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('name_by_lang', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('digest', models.CharField(max_length=64, unique=True)),
                ('related_packages', models.ManyToManyField(related_name='rpm_packagegroup', to='rpm.package')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='PackageEnvironment',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='rpm_packageenvironment', serialize=False, to='core.content')),
                ('id', models.CharField(max_length=255)),
                ('name', models.CharField(max_length=255)),
                ('description', models.TextField(default='')),
                ('display_order', models.IntegerField(null=True)),
                ('group_ids', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('option_ids', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('desc_by_lang', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('name_by_lang', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('digest', models.CharField(max_length=64, unique=True)),
                ('optionalgroups', models.ManyToManyField(related_name='optionalgroups_to_env', to='rpm.packagegroup')),
                ('packagegroups', models.ManyToManyField(related_name='packagegroups_to_env', to='rpm.packagegroup')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='PackageCategory',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='rpm_packagecategory', serialize=False, to='core.content')),
                ('id', models.CharField(max_length=255)),
                ('name', models.CharField(max_length=255)),
                ('description', models.TextField(default='')),
                ('display_order', models.IntegerField(null=True)),
                ('group_ids', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('desc_by_lang', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('name_by_lang', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('digest', models.CharField(max_length=64, unique=True)),
                ('packagegroups', models.ManyToManyField(related_name='rpm_packagecategory', to='rpm.packagegroup')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='Variant',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('variant_id', models.CharField(max_length=50)),
                ('uid', models.CharField(max_length=50)),
                ('name', models.CharField(max_length=50)),
                ('type', models.CharField(max_length=20)),
                ('packages', models.CharField(max_length=50)),
                ('source_packages', models.CharField(max_length=50, null=True)),
                ('source_repository', models.CharField(max_length=50, null=True)),
                ('debug_packages', models.CharField(max_length=50, null=True)),
                ('debug_repository', models.CharField(max_length=50, null=True)),
                ('identity', models.CharField(max_length=50, null=True)),
                ('distribution_tree', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='variants', to='rpm.distributiontree')),
                ('repository', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='+', to='core.repository')),
            ],
            options={
                'unique_together': {('variant_id', 'uid', 'name', 'type', 'packages', 'distribution_tree')},
            },
        ),
        migrations.CreateModel(
            name='Modulemd',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='rpm_modulemd', serialize=False, to='core.content')),
                ('name', models.CharField(max_length=255)),
                ('stream', models.CharField(max_length=255)),
                ('version', models.CharField(max_length=255)),
                ('context', models.CharField(max_length=255)),
                ('arch', models.CharField(max_length=255)),
                ('dependencies', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('artifacts', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('packages', models.ManyToManyField(related_name='rpm_modulemd', to='rpm.package')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('name', 'stream', 'version', 'context', 'arch')},
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='Image',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('name', models.CharField(max_length=20)),
                ('path', models.CharField(max_length=128)),
                ('platforms', models.CharField(max_length=20)),
                ('distribution_tree', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='images', to='rpm.distributiontree')),
            ],
            options={
                'unique_together': {('name', 'path', 'platforms', 'distribution_tree')},
            },
        ),
        migrations.CreateModel(
            name='Checksum',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('path', models.CharField(max_length=128)),
                ('checksum', models.CharField(max_length=128, null=True)),
                ('distribution_tree', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='checksums', to='rpm.distributiontree')),
            ],
            options={
                'unique_together': {('path', 'checksum', 'distribution_tree')},
            },
        ),
        migrations.CreateModel(
            name='Addon',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('addon_id', models.CharField(max_length=50)),
                ('uid', models.CharField(max_length=50)),
                ('name', models.CharField(max_length=50)),
                ('type', models.CharField(max_length=20)),
                ('packages', models.CharField(max_length=50)),
                ('distribution_tree', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='addons', to='rpm.distributiontree')),
                ('repository', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='addons', to='core.repository')),
            ],
            options={
                'unique_together': {('addon_id', 'uid', 'name', 'type', 'packages', 'distribution_tree')},
            },
        ),
    ]



# Generated by Django 4.2.17 on 2024-12-05 18:54

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0003_DATA_incorrect_json'),
    ]

    operations = [
        migrations.AddField(
            model_name='rpmrepository',
            name='metadata_signing_service',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='rpm_rpmrepository', to='core.asciiarmoreddetachedsigningservice'),
        ),
    ]



# Generated by Django 3.2.15 on 2022-08-29 16:33

from django.db import migrations, models

import ast
import logging
import yaml


logger = logging.getLogger(__name__)


def populate_new_fields(apps, schema_editor):
    """Populate new fields of modulemd."""
    Modulemd = apps.get_model("rpm", "Modulemd")
    ModulemdDefaults = apps.get_model("rpm", "ModulemdDefaults")
    modules_to_update = []

    # Fix issue #2786 if already tried to update to 3.18.1
    modulemds_to_update = []
    for modulemd in Modulemd.objects.filter(snippet__startswith="b\'---"):
        modulemd.snippet = ast.literal_eval(modulemd.snippet).decode("utf8")
        modulemds_to_update.append(modulemd)
    Modulemd.objects.bulk_update(modulemds_to_update, ["snippet"])

    # Same issue (#2786) has happened to modulemd defaults
    modulemd_defaults_to_update = []
    for default in ModulemdDefaults.objects.filter(snippet__startswith="b\'---"):
        default.snippet = ast.literal_eval(default.snippet).decode("utf8")
        modulemd_defaults_to_update.append(default)
    ModulemdDefaults.objects.bulk_update(modulemd_defaults_to_update, ["snippet"])

    for modulemd in Modulemd.objects.filter(profiles={}):
        try:
            modulemd_dict = yaml.safe_load(modulemd.snippet)
            modulemd.profiles = modulemd_dict["data"].get("profiles", {})
            modulemd.description = modulemd_dict["data"].get("description", "")
            modules_to_update.append(modulemd)
        except ValueError as err:
            # Due to issue #2735 it could happen that snippet will be empty
            # https://github.com/pulp/pulp_rpm/issues/2735
            logger.warning(
                "Modulemd {}-{}-{} cannot populate new fields."
                "So it will miss the info about profiles and its description.".format(
                    modulemd.name, modulemd.stream, modulemd.version
                )
            )

    Modulemd.objects.bulk_update(modules_to_update, ["profiles", "description"])


class Migration(migrations.Migration):

    dependencies = [
        ("rpm", "0044_noartifact_modules"),
    ]

    operations = [
        migrations.AddField(
            model_name="modulemd",
            name="description",
            field=models.TextField(default="Description"),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="modulemd",
            name="profiles",
            field=models.JSONField(default=dict),
        ),
        migrations.RunPython(populate_new_fields),
    ]



# Generated by Django 3.2.16 on 2022-11-08 19:12

from django.db import migrations, models

import yaml


def parse_date(apps, schema_editor):
    """Parse dates from snippet."""
    ModulemdObsolete = apps.get_model("rpm", "ModulemdObsolete")

    obsoletes_to_update = []

    for obsolete in ModulemdObsolete.objects.all():
        parsed_snippet = yaml.safe_load(obsolete.snippet)
        obsolete.modified = parsed_snippet['data']['modified']
        if parsed_snippet['data'].get('eol_date'):
            obsolete.eol_date = parsed_snippet['data'].get('eol_date')
        obsoletes_to_update.append(obsolete)

    ModulemdObsolete.objects.bulk_update(obsoletes_to_update, ["modified", "eol_date"])


def clean_date_fields(apps, schema_editor):
    """Emtpy date fields."""
    ModulemdObsolete = apps.get_model("rpm", "ModulemdObsolete")
    obsoletes_to_update = []

    for obsolete in ModulemdObsolete.objects.all():
        obsolete.modified = None
        obsolete.eol_date = None

    ModulemdObsolete.objects.bulk_update(obsoletes_to_update, ["modified", "eol_date"])


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0046_rbac_perms'),
    ]

    operations = [
        migrations.AlterField(
            model_name='modulemdobsolete',
            name='modified',
            field=models.DateTimeField(null=True),
        ),
        migrations.AlterField(
            model_name='modulemdobsolete',
            name='eol_date',
            field=models.DateTimeField(null=True),
        ),
        migrations.RunPython(clean_date_fields),
        migrations.AlterField(
            model_name='modulemdobsolete',
            name='eol_date',
            field=models.TextField(null=True),
        ),
        migrations.AlterField(
            model_name='modulemdobsolete',
            name='modified',
            field=models.TextField(),
        ),
        migrations.RunPython(parse_date),
    ]



"""Tests that publish rpm plugin repositories."""

from datetime import datetime
from tempfile import NamedTemporaryFile
from urllib.parse import urljoin

import pytest
import requests
from html.parser import HTMLParser
from productmd.treeinfo import TreeInfo

from pulp_rpm.tests.functional.constants import (
    RPM_PACKAGE_CONTENT_NAME,
    CENTOS8_STREAM_APPSTREAM_URL,
    CENTOS8_STREAM_BASEOS_URL,
)


class PackagesHtmlParser(HTMLParser):
    """HTML parser that looks for the first link to a file.

    This parser look for the first link that doesn't have a trailing slash. It then sets the
    package_href attribute to the value of the href.
    """

    package_href = None

    def handle_starttag(self, tag, attrs):
        """Looks for a tags that don't end with a slash."""
        if not self.package_href:
            if tag == "a" and not attrs[0][1].endswith("/"):
                self.package_href = attrs[0][1]
        else:
            super().handle_starttag(tag, attrs)


def parse_date_from_string(s, parse_format="%Y-%m-%dT%H:%M:%S.%fZ"):
    """Parse string to datetime object.

    :param s: str like '2018-11-18T21:03:32.493697Z'
    :param parse_format: str defaults to %Y-%m-%dT%H:%M:%S.%fZ
    :return: datetime.datetime
    """
    if isinstance(s, datetime):
        return s
    else:
        return datetime.strptime(s, parse_format)


@pytest.fixture
def centos_8stream_baseos_extra_tests(
    rpm_distribution_factory,
    distribution_base_url,
):
    def _extra_test(publication_href):
        # Test that the .treeinfo file is available and AppStream sub-repo is published correctly
        distribution = rpm_distribution_factory(publication=publication_href)
        treeinfo_file = requests.get(
            urljoin(distribution_base_url(distribution.base_url), ".treeinfo")
        ).content
        treeinfo = TreeInfo()
        with NamedTemporaryFile("wb") as temp_file:
            temp_file.write(treeinfo_file)
            temp_file.flush()
            treeinfo.load(f=temp_file.name)
        for variant_name, variant in treeinfo.variants.variants.items():
            if variant_name == "BaseOS":
                assert variant.paths.repository == "."
                assert variant.paths.packages == "Packages"
            elif variant_name == "AppStream":
                assert variant.paths.repository == "AppStream"
                assert variant.paths.packages == "AppStream/Packages"
                # Find the first package in the 'AppStream/Packages/a/' directory and download it
                parser = PackagesHtmlParser()
                a_packages_href = urljoin(
                    distribution_base_url(distribution.base_url),
                    "{}/a/".format(variant.paths.packages),
                )
                a_packages_listing = requests.get(a_packages_href).text
                parser.feed(a_packages_listing)
                full_package_path = urljoin(a_packages_href, parser.package_href)
                assert requests.get(full_package_path).status_code == 200

    return _extra_test


@pytest.mark.parametrize(
    "url,extra_tests",
    [
        (CENTOS8_STREAM_BASEOS_URL, "centos_8stream_baseos_extra_tests"),
        (CENTOS8_STREAM_APPSTREAM_URL, None),
        (CENTOS8_STREAM_BASEOS_URL, None),
    ],
)
def test_rpm_publish(
    url,
    extra_tests,
    init_and_sync,
    rpm_repository_version_api,
    rpm_publication_api,
    monitor_task,
    delete_orphans_pre,
    request,
):
    """Publish repositories with the rpm plugin.

    Do the following:

    1. Create a repository and a remote.
    2. Assert that repository version is None.
    3. Sync the remote.
    4. Assert that repository version is not None.
    5. Assert that distribution_tree units were added and are present in the repo.
    6. Publish
    """
    repo, remote, task = init_and_sync(url=url, policy="on_demand", return_task=True)
    created_at = parse_date_from_string(task.pulp_created)
    started_at = parse_date_from_string(task.started_at)
    finished_at = parse_date_from_string(task.finished_at)
    task_duration = finished_at - started_at
    waiting_time = started_at - created_at
    print(
        "\n->     Sync => Waiting time (s): {wait} | Service time (s): {service}".format(
            wait=waiting_time.total_seconds(), service=task_duration.total_seconds()
        )
    )

    # Check that we have the correct content counts.
    repo_ver = rpm_repository_version_api.read(repo.latest_version_href)
    assert RPM_PACKAGE_CONTENT_NAME in repo_ver.content_summary.present.keys()
    assert RPM_PACKAGE_CONTENT_NAME in repo_ver.content_summary.added.keys()

    # Publishing
    response = rpm_publication_api.create({"repository": repo.pulp_href})
    task = monitor_task(response.task)
    created_at = parse_date_from_string(task.pulp_created)
    started_at = parse_date_from_string(task.started_at)
    finished_at = parse_date_from_string(task.finished_at)
    task_duration = finished_at - started_at
    waiting_time = started_at - created_at
    print(
        "\n->     Publish => Waiting time (s): {wait} | Service time (s): {service}".format(
            wait=waiting_time.total_seconds(), service=task_duration.total_seconds()
        )
    )
    if extra_tests:
        # fixtures and parameterization don't like each other - this hack gets us around that
        xtra_fn = request.getfixturevalue(extra_tests)
        xtra_fn(task.created_resources[0])



from logging import getLogger

from pulpcore.plugin.models import RepositoryVersion
from pulpcore.plugin.viewsets import NamedModelViewSet

from pulp_rpm.app.models.repository import RpmRepository

_logger = getLogger(__name__)


def has_perms_to_copy(request, view, action):
    """
    Check if the source and destination repository matches the usernames permissions.

    `Fail` the check at first missing permission.
    """
    serializer = view.serializer_class(data=request.data, context={"request": request})
    serializer.is_valid(raise_exception=True)

    for copy_action in serializer.data["config"]:
        dest_repo = NamedModelViewSet().get_resource(copy_action["dest_repo"], RpmRepository)

        # Check if user has permissions to destination repository
        if not (
            request.user.has_perm("rpm.modify_content_rpmrepository", dest_repo)
            or request.user.has_perm("rpm.modify_content_rpmrepository")
        ):
            return False

        # Get source repo object
        source_version = NamedModelViewSet().get_resource(
            copy_action["source_repo_version"], RepositoryVersion
        )

        source_repo = RpmRepository.objects.get(pk=source_version.repository_id)

        # Check if user has permissions to source repository
        if not (
            request.user.has_perm("rpm.view_rpmrepository", source_repo)
            or request.user.has_perm("rpm.view_rpmrepository")
        ):
            return False

    return True



# Generated by Django 4.2.17 on 2024-12-05 18:54

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0039_disttree_digest'),
    ]

    operations = [
        migrations.CreateModel(
            name='RpmAlternateContentSource',
            fields=[
                ('alternatecontentsource_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='rpm_rpmalternatecontentsource', serialize=False, to='core.alternatecontentsource')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.alternatecontentsource',),
        ),
    ]



import createrepo_c as cr
import hashlib
import logging
import os
import tempfile
import yaml
import collections

from jsonschema import Draft7Validator
from gettext import gettext as _  # noqa:F401

from pulp_rpm.app.models import Modulemd, Package
from pulp_rpm.app.constants import (
    PULP_MODULEDEFAULTS_ATTR,
    PULP_MODULEOBSOLETES_ATTR,
    PULP_MODULE_ATTR,
    YAML_MODULEMD_OBSOLETES_REQUIRED_ATTR,
    YAML_MODULEMD_DEFAULTS_REQUIRED_ATTR,
)
from pulp_rpm.app.schema import MODULEMD_SCHEMA

log = logging.getLogger(__name__)


def resolve_module_packages(version, previous_version):
    """
    Decide which packages to add/remove based on modular data.

    Args:
        version (pulpcore.app.models.RepositoryVersion): current incomplete repository version
        previous_version (pulpcore.app.models.RepositoryVersion) :  previous version of the same
                                                                    repository to compare to

    """

    def modules_packages(modules):
        packages = set()
        for module in modules:
            packages.update(module.packages.all().only("pk"))
        return packages

    modulemd_pulp_type = Modulemd.get_pulp_type()
    current_modules = Modulemd.objects.filter(
        pk__in=version.content.filter(pulp_type=modulemd_pulp_type)
    )
    current_module_packages = modules_packages(current_modules)

    if previous_version:
        previous_modules = Modulemd.objects.filter(
            pk__in=previous_version.content.filter(pulp_type=modulemd_pulp_type)
        )
        added_modules = current_modules.difference(previous_modules)
        removed_modules = previous_modules.difference(current_modules)
        removed_module_packages = modules_packages(removed_modules)
        packages_to_remove = removed_module_packages.difference(current_module_packages)
        version.remove_content(Package.objects.filter(pk__in=packages_to_remove))
    else:
        added_modules = current_modules

    added_module_packages = modules_packages(added_modules)
    packages_to_add = added_module_packages.difference(current_module_packages)
    version.add_content(Package.objects.filter(pk__in=packages_to_add))


def split_modulemd_file(file: str):
    """
    Helper method to preserve original formatting of modulemd.

    Args:
        file: Absolute path to file
    """
    with tempfile.TemporaryDirectory(dir=".") as tf:
        decompressed_path = os.path.join(tf, "modulemd.yaml")
        cr.decompress_file(file, decompressed_path, cr.AUTO_DETECT_COMPRESSION)
        with open(decompressed_path) as modulemd_file:
            for doc in modulemd_file.read().split("---"):
                # strip any spaces or newlines from either side, strip the document end marking,
                # then strip again so we have only the document text w/o newlines
                stripped = doc.strip().rstrip("...").rstrip()
                if stripped:
                    # add the document begin/end markers backs
                    normalized = "---\n{}\n...".format(stripped)
                    yield normalized


def check_mandatory_module_fields(module, required_fields):
    """
    Check mandatory fields on module dict.
    """
    data = module["data"]
    for field in required_fields:
        if field not in data.keys():
            raise ValueError(
                _("Mandatory field {} is missing in {}.".format(field, module["document"]))
            )


def create_modulemd(modulemd, snippet):
    """
    Create dict with modulemd data to be saved to DB.
    """
    new_module = dict()
    new_module[PULP_MODULE_ATTR.NAME] = modulemd["data"].get("name")
    new_module[PULP_MODULE_ATTR.STREAM] = str(modulemd["data"].get("stream"))
    new_module[PULP_MODULE_ATTR.VERSION] = str(modulemd["data"].get("version"))
    new_module[PULP_MODULE_ATTR.STATIC_CONTEXT] = modulemd["data"].get("static_context")
    new_module[PULP_MODULE_ATTR.CONTEXT] = modulemd["data"].get("context")
    new_module[PULP_MODULE_ATTR.ARCH] = modulemd["data"].get("arch")
    new_module[PULP_MODULE_ATTR.ARTIFACTS] = modulemd["data"].get("artifacts", {}).get("rpms", [])
    new_module[PULP_MODULE_ATTR.DESCRIPTION] = modulemd["data"].get("description")
    new_module[PULP_MODULE_ATTR.DEPENDENCIES] = modulemd["data"].get("dependencies", [])

    # keep data formatted the same as it was with previous parsing implementation
    unprocessed_profiles = modulemd["data"].get("profiles", {})
    profiles = {}
    if unprocessed_profiles:
        for name, data in unprocessed_profiles.items():
            rpms = data.get("rpms")
            if not rpms:
                msg = (
                    "Got unexpected data for module {}-{}-{}-{}-{}: "
                    "profiles failed to parse properly"
                ).format(
                    new_module[PULP_MODULE_ATTR.NAME],
                    new_module[PULP_MODULE_ATTR.STREAM],
                    new_module[PULP_MODULE_ATTR.VERSION],
                    new_module[PULP_MODULE_ATTR.CONTEXT],
                    new_module[PULP_MODULE_ATTR.ARCH],
                )
                log.warning(msg)
            else:
                profiles[name] = rpms

    new_module[PULP_MODULE_ATTR.PROFILES] = profiles
    new_module["snippet"] = snippet
    new_module["digest"] = hashlib.sha256(snippet.encode()).hexdigest()

    return new_module


def create_modulemd_defaults(default, snippet):
    """
    Create dict with modulemd-defaults data to can be saved to DB.
    """
    new_default = dict()
    new_default[PULP_MODULEDEFAULTS_ATTR.MODULE] = default["data"].get("module")
    new_default[PULP_MODULEDEFAULTS_ATTR.STREAM] = str(default["data"].get("stream", ""))
    new_default[PULP_MODULEDEFAULTS_ATTR.PROFILES] = default["data"].get("profiles")
    new_default["snippet"] = snippet
    new_default[PULP_MODULEDEFAULTS_ATTR.DIGEST] = hashlib.sha256(snippet.encode()).hexdigest()

    return new_default


def create_modulemd_obsoletes(obsolete, snippet):
    """
    Create dict with modulemd-obsoletes data to can be saved to DB.
    """
    new_obsolete = dict()

    new_obsolete[PULP_MODULEOBSOLETES_ATTR.MODIFIED] = obsolete["data"].get("modified")
    new_obsolete[PULP_MODULEOBSOLETES_ATTR.MODULE] = obsolete["data"].get("module")
    new_obsolete[PULP_MODULEOBSOLETES_ATTR.STREAM] = str(obsolete["data"].get("stream"))
    new_obsolete[PULP_MODULEOBSOLETES_ATTR.MESSAGE] = obsolete["data"].get("message")
    new_obsolete[PULP_MODULEOBSOLETES_ATTR.RESET] = obsolete["data"].get("reset")
    new_obsolete[PULP_MODULEOBSOLETES_ATTR.CONTEXT] = obsolete["data"].get("context")

    if obsolete["data"].get("eol_date"):
        new_obsolete[PULP_MODULEOBSOLETES_ATTR.EOL] = obsolete["data"].get("eol_date")
    if obsolete["data"].get("obsoleted_by"):
        new_obsolete[PULP_MODULEOBSOLETES_ATTR.OBSOLETE_BY_MODULE] = obsolete["data"][
            "obsoleted_by"
        ].get("module")
        new_obsolete[PULP_MODULEOBSOLETES_ATTR.OBSOLETE_BY_STREAM] = obsolete["data"][
            "obsoleted_by"
        ].get("stream")
    new_obsolete["snippet"] = snippet

    return new_obsolete


def parse_modular(file: str):
    """
    Parse all modular metadata.

    Args:
        file: Absolute path to file
    """
    modulemd_all = []
    modulemd_defaults_all = []
    modulemd_obsoletes_all = []

    for module in split_modulemd_file(file):
        parsed_data = yaml.load(module, Loader=ModularYamlLoader)
        # here we check the modulemd document as we don't store all info, so serializers
        # are not enough then we only need to take required data from dict which is
        # parsed by pyyaml library
        if parsed_data["document"] == "modulemd":
            # the validator currently accepts formatting slightly different to the
            # spec due to the misconfiguration of some Rocky Linux 9 repositories
            # https://bugs.rockylinux.org/view.php?id=2575
            # further discussion on this issue can be found here:
            # https://github.com/pulp/pulp_rpm/issues/2998
            validator = Draft7Validator(MODULEMD_SCHEMA)
            err = []
            for error in sorted(validator.iter_errors(parsed_data["data"]), key=str):
                err.append(error.message)
            if err:
                raise ValueError(_("Provided modular data is invalid:'{}'".format(err)))
            modulemd_all.append(create_modulemd(parsed_data, module))
        elif parsed_data["document"] == "modulemd-defaults":
            check_mandatory_module_fields(parsed_data, YAML_MODULEMD_DEFAULTS_REQUIRED_ATTR)
            modulemd_defaults_all.append(create_modulemd_defaults(parsed_data, module))
        elif parsed_data["document"] == "modulemd-obsoletes":
            check_mandatory_module_fields(parsed_data, YAML_MODULEMD_OBSOLETES_REQUIRED_ATTR)
            modulemd_obsoletes_all.append(create_modulemd_obsoletes(parsed_data, module))
        else:
            logging.warning(f"Unknown modular document type found: {parsed_data.get('document')}")

    return modulemd_all, modulemd_defaults_all, modulemd_obsoletes_all


class ModularYamlLoader(yaml.SafeLoader):
    """
    Custom Loader that preserve unquoted float in specific fields (see #3285).

    Motivation (for customizing YAML parsing) is that libmodulemd also implement safe-quoting:
    https://github.com/fedora-modularity/libmodulemd/blob/main/modulemd/tests/test-modulemd-quoting.c

    This class is based on https://stackoverflow.com/a/74334992
    """

    # Field to preserve (will bypass yaml casting)
    PRESERVED_FIELDS = ("name", "stream", "version", "context", "arch")

    def construct_mapping(self, node, deep=False):
        if not isinstance(node, yaml.MappingNode):
            raise yaml.constructor.ConstructorError(
                None, None, "expected a mapping node, but found %s" % node.id, node.start_mark
            )
        mapping = {}
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            if not isinstance(key, collections.abc.Hashable):
                raise yaml.constructor.ConstructorError(
                    "while constructing a mapping",
                    node.start_mark,
                    "found unhashable key",
                    key_node.start_mark,
                )
            if key in ModularYamlLoader.PRESERVED_FIELDS and isinstance(
                value_node, yaml.ScalarNode
            ):
                value = value_node.value
            else:
                value = self.construct_object(value_node, deep=deep)
            mapping[key] = value
        return mapping



from tempfile import NamedTemporaryFile

from pulpcore.plugin.models import Artifact, CreatedResource, PulpTemporaryFile, Upload, UploadChunk
from pulpcore.plugin.tasking import general_create
from pulpcore.plugin.util import get_url

from pulp_rpm.app.models.content import RpmPackageSigningService


def _save_file(fileobj, final_package):
    with fileobj.file.open() as fd:
        final_package.write(fd.read())
    final_package.flush()


def _save_upload(uploadobj, final_package):
    chunks = UploadChunk.objects.filter(upload=uploadobj).order_by("offset")
    for chunk in chunks:
        final_package.write(chunk.file.read())
        chunk.file.close()
    final_package.flush()


def sign_and_create(
    app_label,
    serializer_name,
    signing_service_pk,
    signing_fingerprint,
    temporary_file_pk,
    *args,
    **kwargs,
):
    data = kwargs.pop("data", None)
    context = kwargs.pop("context", {})

    # Get unsigned package file and sign it
    package_signing_service = RpmPackageSigningService.objects.get(pk=signing_service_pk)
    with NamedTemporaryFile(mode="wb", dir=".", delete=False) as final_package:
        try:
            uploaded_package = PulpTemporaryFile.objects.get(pk=temporary_file_pk)
            _save_file(uploaded_package, final_package)
        except PulpTemporaryFile.DoesNotExist:
            uploaded_package = Upload.objects.get(pk=temporary_file_pk)
            _save_upload(uploaded_package, final_package)

        package_signing_service.sign(final_package.name, pubkey_fingerprint=signing_fingerprint)
        artifact = Artifact.init_and_validate(final_package.name)
        artifact.save()
        resource = CreatedResource(content_object=artifact)
        resource.save()
    uploaded_package.delete()

    # Create Package content
    data["artifact"] = get_url(artifact)
    # The Package serializer validation method have two branches: the signing and non-signing.
    # Here, the package is already signed, so we need to update the context for a proper validation.
    context["sign_package"] = False
    # The request data is immutable when there's an upload, so we can't delete the upload out of the
    # request data like we do for a file.  Instead, we'll delete it here.
    if "upload" in data:
        del data["upload"]
    general_create(app_label, serializer_name, data=data, context=context, *args, **kwargs)



import tempfile
from pathlib import Path

from django.conf import settings
from pulpcore.plugin.exceptions import PulpException
from pulpcore.plugin.models import SigningService
from typing import Optional

from pulp_rpm.app.shared_utils import RpmTool


class RpmPackageSigningService(SigningService):
    """
    A model used for signing RPM packages.

    The pubkey_fingerprint should be passed explicitly in the sign method.
    """

    def _env_variables(self, env_vars=None):
        # Prevent the signing service pubkey to be used for signing a package.
        # The pubkey should be provided explicitly.
        _env_vars = {"PULP_SIGNING_KEY_FINGERPRINT": None}
        if env_vars:
            _env_vars.update(env_vars)
        return super()._env_variables(_env_vars)

    def sign(
        self,
        filename: str,
        env_vars: Optional[dict] = None,
        pubkey_fingerprint: Optional[str] = None,
    ):
        """
        Sign a package @filename using @pubkey_figerprint.

        Args:
            filename: The absolute path to the package to be signed.
            env_vars: (optional) Dict of env_vars to be passed to the signing script.
            pubkey_fingerprint: The V4 fingerprint that correlates with the private key to use.
        """
        if not pubkey_fingerprint:
            raise ValueError("A pubkey_fingerprint must be provided.")
        _env_vars = env_vars or {}
        _env_vars["PULP_SIGNING_KEY_FINGERPRINT"] = pubkey_fingerprint
        return super().sign(filename, _env_vars)

    def validate(self):
        """
        Validate a signing service for a Rpm Package signature.

        Specifically, it validates that self.signing_script can sign an rpm package with
        the sample key self.pubkey and that the self.sign() method returns:

        ```json
        {"rpm_package": "<path/to/package.rpm>"}
        ```

        See [RpmTool.verify_signature][] for the signature verificaton method used.
        """
        with tempfile.TemporaryDirectory(dir=settings.WORKING_DIRECTORY) as temp_directory_name:
            # get and sign sample rpm
            temp_file = RpmTool.get_empty_rpm(temp_directory_name)
            return_value = self.sign(temp_file, pubkey_fingerprint=self.pubkey_fingerprint)
            try:
                return_value["rpm_package"]
            except KeyError:
                raise PulpException(f"Malformed output from signing script: {return_value}")

            # verify with rpm tool
            rpm_tool = RpmTool(root=Path(temp_directory_name))
            rpm_tool.import_pubkey_string(self.public_key)
            rpm_tool.verify_signature(temp_file)



# Generated by Django 2.2.24 on 2021-08-03 21:04

from django.db import migrations


def migrate_sub_repo_value(apps, schema_editor):
    RpmRepository = apps.get_model('rpm', 'RpmRepository')
    RpmRepository.objects.filter(sub_repo=True).update(user_hidden=True)


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0036_checksum_type'),
    ]

    operations = [
        migrations.RunPython(
            migrate_sub_repo_value
        ),
        migrations.RemoveField(
            model_name='rpmrepository',
            name='sub_repo',
        ),
    ]



# Generated by Django 4.2.7 on 2023-11-29 23:12

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0114_remove_task_args_remove_task_kwargs"),
        ("rpm", "0057_rpmpublication_checksum_type_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="addon",
            name="repository",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.RESTRICT,
                related_name="addons",
                to="core.repository",
            ),
        ),
        migrations.AlterField(
            model_name="variant",
            name="repository",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.RESTRICT,
                related_name="variants",
                to="core.repository",
            ),
        ),
    ]



# Generated by Django 2.2.14 on 2020-08-06 00:50

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0019_migrate_updatecollection_data'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='updatecollection',
            name='update_record',
        ),
    ]
